"""This module contains transformers that apply encodings to nominal columns."""
from __future__ import annotations

import warnings
from typing import Literal

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from tubular.base import BaseTransformer
from tubular.mapping import BaseMappingTransformMixin


class BaseNominalTransformer(BaseTransformer):
    """Base Transformer extension for nominal transformers.

    Contains columns_set_or_check method which overrides the columns_set_or_check method in BaseTransformer if given
    primacy in inheritance. The difference being that BaseNominalTransformer's columns_set_or_check only selects
    object and categorical columns from X, if the columns attribute is not set by the user.
    """

    def columns_set_or_check(self, X: pd.DataFrame) -> None:
        """Function to check or set columns attribute.

        If the columns attribute is None then set it to all object and category columns in X. Otherwise run the
        columns_check method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to check columns are in.

        """
        if self.columns is None:
            columns = [
                c for c in X.columns if X[c].dtype.name in ["object", "category"]
            ]

            if not len(columns) > 0:
                msg = f"{self.classname()}: no object or category columns in X"
                raise ValueError(msg)

            self.columns = columns

        else:
            self.columns_check(X)

    def check_mappable_rows(self, X: pd.DataFrame) -> None:
        """Method to check that all the rows to apply the transformer to are able to be
        mapped according to the values in the mappings dict.

        Raises
        ------
        ValueError
            If any of the rows in a column (c) to be mapped, could not be mapped according to
            the mapping dict in mappings[c].

        """
        self.check_is_fitted(["mappings"])

        for c in self.columns:
            mappable_rows = X[c].isin(list(self.mappings[c])).sum()

            if mappable_rows < X.shape[0]:
                msg = f"{self.classname()}: nulls would be introduced into column {c} from levels not present in mapping"
                raise ValueError(msg)


class NominalToIntegerTransformer(BaseNominalTransformer, BaseMappingTransformMixin):
    """Transformer to convert columns containing nominal values into integer values.

    The nominal levels that are mapped to integers are not ordered in any way.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    start_encoding : int, default = 0
        Value to start the encoding from e.g. if start_encoding = 0 then the encoding would be
        {'A': 0, 'B': 1, 'C': 3} etc.. or if start_encoding = 5 then the same encoding would be
        {'A': 5, 'B': 6, 'C': 7}. Can be positive or negative.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    start_encoding : int
        Value to start the encoding / mapping of nominal to integer from.

    mappings : dict
        Created in fit. A dict of key (column names) value (mappings between levels and integers for given
        column) pairs.

    inverse_mapping_ : dict
        Created in inverse_transform. Inverse mapping of mappings. Maps integer value back to categorical
        levels.

    """

    def __init__(
        self,
        columns: str | list[str] | None = None,
        start_encoding: int = 0,
        **kwargs: dict[str, bool],
    ) -> None:
        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

        if not isinstance(start_encoding, int):
            msg = f"{self.classname()}: start_encoding should be an integer"
            raise ValueError(msg)

        self.start_encoding = start_encoding

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Creates mapping between nominal levels and integer values for categorical variables.

        Parameters
        ----------
        X : pd.DataFrame
            Data to fit the transformer on, this sets the nominal levels that can be mapped.

        y : None or pd.DataFrame or pd.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}

        for c in self.columns:
            col_values = X[c].unique()

            self.mappings[c] = {
                k: i for i, k in enumerate(col_values, self.start_encoding)
            }

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform method to apply integer encoding stored in the mappings attribute to
        each column in the columns attribute.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        Parameters
        ----------
        X : pd.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """

        super(BaseNominalTransformer, self).columns_check(X)

        self.check_mappable_rows(X)

        return BaseMappingTransformMixin.transform(self, X)

    def inverse_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Converts integer values back to categorical / nominal values. Does the inverse of the transform method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with integer columns to convert back to catgeorical.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with integers mapped back to categorical levels.

        """
        X = BaseNominalTransformer.transform(self, X)

        self.check_is_fitted(["mappings"])

        self.inverse_mapping_ = {}

        for c in self.columns:
            # calculate the reverse mapping
            self.inverse_mapping_[c] = {v: k for k, v in self.mappings[c].items()}

            mappable_rows = (
                X[c].isin([k for k, v in self.inverse_mapping_[c].items()]).sum()
            )

            X[c] = X[c].replace(self.inverse_mapping_[c])

            if (X.shape[0] - mappable_rows) > 0:
                raise ValueError(
                    f"{self.classname()}: nulls introduced from levels not present in mapping for column: "
                    + c,
                )

        return X


class GroupRareLevelsTransformer(BaseNominalTransformer):
    """Transformer to group together rare levels of nominal variables into a new level,
    labelled 'rare' (by default).

    Rare levels are defined by a cut off percentage, which can either be based on the
    number of rows or sum of weights. Any levels below this cut off value will be
    grouped into the rare level.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    cut_off_percent : float, default = 0.01
        Cut off for the percent of rows or percent of weight for a level, levels below
        this value will be grouped.

    weight : None or str, default = None
        Name of weights column that should be used so cut_off_percent applies to sum of weights
        rather than number of rows.

    rare_level_name : any,default = 'rare'.
        Must be of the same type as columns.
        Label for the new 'rare' level.

    record_rare_levels : bool, default = False
        If True, an attribute called rare_levels_record_ will be added to the object. This will be a dict
        of key (column name) value (level from column considered rare according to cut_off_percent) pairs.
        Care should be taken if working with nominal variables with many levels as this could potentially
        result in many being stored in this attribute.

    unseen_levels_to_rare : bool, default = True
        If True, unseen levels in new data will be passed to rare, if set to false they will be left unchanged.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    cut_off_percent : float
        Cut off percentage (either in terms of number of rows or sum of weight) for a given
        nominal level to be considered rare.

    non_rare_levels : dict
        Created in fit. A dict of non-rare levels (i.e. levels with more than cut_off_percent weight or rows)
        that is used to identify rare levels in transform.

    rare_level_name : any
        Must be of the same type as columns.
        Label for the new nominal level that will be added to group together rare levels (as
        defined by cut_off_percent).

    record_rare_levels : bool
        Should the 'rare' levels that will be grouped together be recorded? If not they will be lost
        after the fit and the only information remaining will be the 'non'rare' levels.

    rare_levels_record_ : dict
        Only created (in fit) if record_rare_levels is True. This is dict containing a list of
        levels that were grouped into 'rare' for each column the transformer was applied to.

    weight : str
        Name of weights columns to use if cut_off_percent should be in terms of sum of weight
        not number of rows.

    unseen_levels_to_rare : bool
        If True, unseen levels in new data will be passed to rare, if set to false they will be left unchanged.

    training_data_levels : dict[set]
        Dictionary containing the set of values present in the training data for each column in self.columns. It
        will only exist in if unseen_levels_to_rare is set to False.

    """

    def __init__(
        self,
        columns: str | list[str] | None = None,
        cut_off_percent: float = 0.01,
        weight: str | None = None,
        rare_level_name: str | list[str] | None = "rare",
        record_rare_levels: bool = True,
        unseen_levels_to_rare: bool = True,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if not isinstance(cut_off_percent, float):
            msg = f"{self.classname()}: cut_off_percent must be a float"
            raise ValueError(msg)

        if not ((cut_off_percent > 0) & (cut_off_percent < 1)):
            msg = f"{self.classname()}: cut_off_percent must be > 0 and < 1"
            raise ValueError(msg)

        self.cut_off_percent = cut_off_percent

        if weight is not None and not isinstance(weight, str):
            msg = f"{self.classname()}: weight should be a single column (str)"
            raise ValueError(msg)

        self.weight = weight

        self.rare_level_name = rare_level_name

        if not isinstance(record_rare_levels, bool):
            msg = f"{self.classname()}: record_rare_levels must be a bool"
            raise ValueError(msg)

        self.record_rare_levels = record_rare_levels

        if not isinstance(unseen_levels_to_rare, bool):
            msg = f"{self.classname()}: unseen_levels_to_rare must be a bool"
            raise ValueError(msg)

        self.unseen_levels_to_rare = unseen_levels_to_rare

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Records non-rare levels for categorical variables.

        When transform is called, only levels records in non_rare_levels during fit will remain
        unchanged - all other levels will be grouped. If record_rare_levels is True then the
        rare levels will also be recorded.

        The label for the rare levels must be of the same type as the columns.

        Parameters
        ----------
        X : pd.DataFrame
            Data to identify non-rare levels from.

        y : None or pd.DataFrame or pd.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        super().fit(X, y)

        for c in self.columns:
            if (X[c].dtype.name != "category") and (
                pd.Series(self.rare_level_name).dtype != X[c].dtypes
            ):
                msg = f"{self.classname()}: rare_level_name must be of the same type of the columns"
                raise ValueError(msg)

        if self.weight is not None and self.weight not in X.columns.to_numpy():
            msg = f"{self.classname()}: weight {self.weight} not in X"
            raise ValueError(msg)

        self.non_rare_levels = {}

        if self.record_rare_levels:
            self.rare_levels_record_ = {}

        if self.weight is None:
            for c in self.columns:
                col_percents = X[c].value_counts(dropna=False) / X.shape[0]

                self.non_rare_levels[c] = list(
                    col_percents.loc[col_percents >= self.cut_off_percent].index.values,
                )

                self.non_rare_levels[c] = sorted(self.non_rare_levels[c], key=str)

                if self.record_rare_levels:
                    self.rare_levels_record_[c] = list(
                        col_percents.loc[
                            col_percents < self.cut_off_percent
                        ].index.values,
                    )

                    self.rare_levels_record_[c] = sorted(
                        self.rare_levels_record_[c],
                        key=str,
                    )

        else:
            for c in self.columns:
                cols_w_percents = X.groupby(c)[self.weight].sum()

                # nulls are excluded from pandas groupby; https://github.com/pandas-dev/pandas/issues/3729
                # so add them back in
                if cols_w_percents.sum() < X[self.weight].sum():
                    cols_w_percents[np.NaN] = X.loc[X[c].isna(), self.weight].sum()

                cols_w_percents = cols_w_percents / X[self.weight].sum()

                self.non_rare_levels[c] = list(
                    cols_w_percents.loc[
                        cols_w_percents >= self.cut_off_percent
                    ].index.values,
                )

                self.non_rare_levels[c] = sorted(self.non_rare_levels[c], key=str)

                if self.record_rare_levels:
                    self.rare_levels_record_[c] = list(
                        cols_w_percents.loc[
                            cols_w_percents < self.cut_off_percent
                        ].index.values,
                    )

                    self.rare_levels_record_[c] = sorted(
                        self.rare_levels_record_[c],
                        key=str,
                    )

        if not self.unseen_levels_to_rare:
            self.training_data_levels = {}
            for c in self.columns:
                self.training_data_levels[c] = set(X[c])

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Grouped rare levels together into a new 'rare' level.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variables to apply rare level grouping to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with rare levels grouped for into a new rare level.

        """
        X = BaseNominalTransformer.transform(self, X)

        self.check_is_fitted(["non_rare_levels"])

        if not self.unseen_levels_to_rare:
            for c in self.columns:
                unseen_vals = set(X[c]) - set(self.training_data_levels[c])
                for unseen_val in unseen_vals:
                    self.non_rare_levels[c].append(unseen_val)

        for c in self.columns:
            # for categorical dtypes have to set new category for the impute values first
            # and convert back to the categorical type, other it will convert to object
            if "category" in X[c].dtype.name:

                categories_before = X[c].dtype.categories

                if self.rare_level_name not in X[c].cat.categories:
                    X[c] = X[c].cat.add_categories(self.rare_level_name)

                X[c] = pd.Series(
                    data=np.where(
                        X[c].isin(self.non_rare_levels[c]),
                        X[c],
                        self.rare_level_name,
                    ),
                    index=X.index,
                )

                remaining_categories = [
                    category
                    for category in categories_before
                    if category in self.non_rare_levels[c]
                ]

                X[c] = pd.Categorical(
                    X[c],
                    categories=remaining_categories + [self.rare_level_name],
                )

            else:
                # using np.where converts np.NaN to str value if only one row of data frame is passed
                # instead, using pd.where(), if condition true, keep original value, else replace with self.rare_level_name
                X[c] = X[c].where(
                    X[c].isin(self.non_rare_levels[c]),
                    self.rare_level_name,
                )

        return X


class MeanResponseTransformer(BaseNominalTransformer):
    """Transformer to apply mean response encoding. This converts categorical variables to
    numeric by mapping levels to the mean response for that level.

    For a continuous or binary response the categorical columns specified will have values
    replaced with the mean response for each category.

    For an n > 1 level categorical response, up to n binary responses can be created, which in
    turn can then be used to encode each categorical column specified. This will generate up
    to n * len(columns) new columns, of with names of the form {column}_{response_level}. The
    original columns will be removed from the dataframe. This functionality is controlled using
    the 'level' parameter. Note that the above only works for a n > 1 level categorical response.
    Do not use 'level' parameter for a n = 1 level numerical response. In this case, use the standard
    mean response transformer without the 'level' parameter.

    If a categorical variable contains null values these will not be transformed.

    The same weights and prior are applied to each response level in the multi-level case.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    prior : int, default = 0
        Regularisation parameter, can be thought of roughly as the size a category should be in order for
        its statistics to be considered reliable (hence default value of 0 means no regularisation).

    level : str, list or None, default = None
        Parameter to control encoding against a multi-level categorical response. For a continuous or
        binary response, leave this as None. In the multi-level case, set to 'all' to encode against every
        response level or provide a list of response levels to encode against.

    unseen_level_handling : str("Mean", "Median", "Lowest" or "Highest) or int/float, default = None
        Parameter to control the logic for handling unseen levels of the categorical features to encode in
        data when using transform method. Default value of None will output error when attempting to use transform
        on data with unseen levels in categorical columns to encode. Set this parameter to one of the options above
        in order to encode unseen levels in each categorical column with the mean, median etc. of
        each column. One can also pass an arbitrary int/float value to use for encoding unseen levels.

    return_type: Literal['float32', 'float64']
        What type to cast return column as, consider exploring float32 to save memory. Defaults to float32.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    columns : str or list
        Categorical columns to encode in the input data.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    prior : int, default = 0
        Regularisation parameter, can be thought of roughly as the size a category should be in order for
        its statistics to be considered reliable (hence default value of 0 means no regularisation).

    level : str, list or None, default = None
        Parameter to control encoding against a multi-level categorical response. If None the response will be
        treated as binary or continous, if 'all' all response levels will be encoded against and if it is a list of
        levels then only the levels specified will be encoded against.

    response_levels : list
        Only created in the mutli-level case. Generated from level, list of all the response levels to encode against.

    mappings : dict
        Created in fit. A nested Dict of {column names : column specific mapping dictionary} pairs.  Column
        specific mapping dictionaries contain {initial value : mapped value} pairs.

    mapped_columns : list
        Only created in the multi-level case. A list of the new columns produced by encoded the columns in self.columns
        against multiple response levels, of the form {column}_{level}.

    transformer_dict : dict
        Only created in the mutli-level case. A dictionary of the form level : transformer containing the mean response
        transformers for each level to be encoded against.

    unseen_levels_encoding_dict: dict
        Dict containing the values (based on chosen unseen_level_handling) derived from the encoded columns to use when handling unseen levels in data passed to transform method.

    return_type: Literal['float32', 'float64']
        What type to cast return column as. Defaults to float32.

    cast_method: Literal[np.float32, np,float64]
        Store the casting method associated to return_type

    """

    def __init__(
        self,
        columns: str | list[str] | None = None,
        weights_column: str | None = None,
        prior: int = 0,
        level: str | list | None = None,
        unseen_level_handling: str | int | float | None = None,
        return_type: Literal["float32", "float64"] = "float32",
        **kwargs: dict[str, bool],
    ) -> None:
        if weights_column is not None and type(weights_column) is not str:
            msg = f"{self.classname()}: weights_column should be a str"
            raise TypeError(msg)

        if type(prior) is not int:
            msg = f"{self.classname()}: prior should be a int"
            raise TypeError(msg)

        if not prior >= 0:
            msg = f"{self.classname()}: prior should be positive int"
            raise ValueError(msg)

        if level and not isinstance(level, str) and not isinstance(level, list):
            msg = f"{self.classname()}: Level should be a NoneType, list or str but got {type(level)}"
            raise TypeError(msg)
        if (
            unseen_level_handling
            and (unseen_level_handling not in ["Mean", "Median", "Lowest", "Highest"])
            and not (isinstance(unseen_level_handling, (int, float)))
        ):
            msg = f"{self.classname()}: unseen_level_handling should be the option: Mean, Median, Lowest, Highest or an arbitrary int/float value"
            raise ValueError(msg)

        if return_type not in ["float64", "float32"]:
            msg = f"{self.classname()}: return_type should be one of: 'float64', 'float32'"
            raise ValueError(msg)

        self.weights_column = weights_column
        self.prior = prior
        self.level = level
        self.unseen_level_handling = unseen_level_handling
        self.return_type = return_type
        if return_type == "float64":
            self.cast_method = np.float64
        else:
            self.cast_method = np.float32
        # TODO: set default prior to None and refactor to only use prior regularisation when it is set?

        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

    def _prior_regularisation(
        self,
        target_means: pd.Series,
        cat_freq: str,
    ) -> pd.Series:
        """Regularise encoding values by pushing encodings of infrequent categories towards the global mean.  If prior is zero this will return target_means unaltered.

        Parameters
        ----------
        target_means : pd.Series
            Series containing group means for levels of column in data

        cat_freq : str
            Series containing group sizes for levels of column in data

        Returns
        -------
        regularised : pd.Series
            Series of regularised encoding values
        """
        self.check_is_fitted(["global_mean"])

        return (
            target_means.multiply(cat_freq, axis="index")
            + self.global_mean * self.prior
        ).divide(cat_freq + self.prior, axis="index")

    def _fit_binary_response(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        columns: list[str],
    ) -> None:
        """Function to learn the MRE mappings for a given binary or continuous response.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform.

        y : pd.Series
            Binary or contionuous response variable to encode against.

        columns : list(str)
            Post transform names of columns to be encoded. In the binary or continous case
            this is just self.columns. In the multi-level case this should be of the form
            {column_in_original_data}_{response_level}, where response_level is the level
            being encoded against in this call of _fit_binary_response.
        """
        if (
            self.weights_column is not None
            and self.weights_column not in X.columns.to_numpy()
        ):
            msg = f"{self.classname()}: weights column {self.weights_column} not in X"
            raise ValueError(msg)

        response_null_count = y.isna().sum()

        if response_null_count > 0:
            msg = f"{self.classname()}: y has {response_null_count} null values"
            raise ValueError(msg)

        X_y = self._combine_X_y(X, y)
        response_column = "_temporary_response"

        if self.weights_column is None:
            self.global_mean = X_y[response_column].mean()

        else:
            X_y["weighted_response"] = X_y[response_column].multiply(
                X_y[self.weights_column],
            )

            self.global_mean = (
                X_y["weighted_response"].sum() / X_y[self.weights_column].sum()
            )

        for c in columns:
            if self.weights_column is None:
                group_means = X_y.groupby(c)[response_column].mean()

                group_counts = X_y.groupby(c)[response_column].size()

                self.mappings[c] = self._prior_regularisation(
                    group_means,
                    group_counts,
                ).to_dict()

            else:
                groupby_sum = X_y.groupby([c])[
                    ["weighted_response", self.weights_column]
                ].sum()

                group_weight = groupby_sum[self.weights_column]

                group_means = groupby_sum["weighted_response"] / group_weight

                self.mappings[c] = self._prior_regularisation(
                    group_means,
                    group_weight,
                ).to_dict()

            # to_dict changes types
            for key in self.mappings[c]:
                self.mappings[c][key] = self.cast_method(self.mappings[c][key])

    def fit(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """Identify mapping of categorical levels to mean response values.

        If the user specified the weights_column arg in when initialising the transformer
        the weighted mean response will be calculated using that column.

        In the multi-level case this method learns which response levels are present and
        are to be encoded against.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform and also containing response_column
            column.

        y : pd.Series
            Response variable or target.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}
        self.unseen_levels_encoding_dict = {}
        X_temp = X.copy()

        if self.level:
            if self.level == "all":
                self.response_levels = y.unique()

            else:
                if isinstance(self.level, str):
                    self.level = [self.level]

                if any(level not in list(y.unique()) for level in self.level):
                    msg = "Levels contains a level to encode against that is not present in the response."
                    raise ValueError(msg)

                self.response_levels = self.level

            self.transformer_dict = {}
            mapped_columns = []

            for level in self.response_levels:
                mapping_columns_for_this_level = [
                    column + "_" + level for column in self.columns
                ]

                for column in self.columns:
                    X_temp[column + "_" + level] = X[column].copy()

                # keep nans to preserve null check functionality of binary response MRE transformer
                y_temp = y.apply(
                    lambda x, level=level: x == level if not pd.isna(x) else np.nan,
                )

                self.transformer_dict[level] = self._fit_binary_response(
                    X_temp,
                    y_temp,
                    mapping_columns_for_this_level,
                )

                mapped_columns += mapping_columns_for_this_level

            self.mapped_columns = list(set(mapped_columns) - set(self.columns))
            self.encoded_feature_columns = self.mapped_columns

        else:
            self._fit_binary_response(X, y, self.columns)
            self.encoded_feature_columns = self.columns

        if isinstance(self.unseen_level_handling, (int, float)):
            for c in self.encoded_feature_columns:
                self.unseen_levels_encoding_dict[c] = self.cast_method(
                    self.unseen_level_handling,
                )

        else:
            for c in self.encoded_feature_columns:
                X_temp[c] = X_temp[c].map(self.mappings[c]).astype(self.return_type)

                if self.unseen_level_handling == "Mean":
                    self.unseen_levels_encoding_dict[c] = X_temp[c].mean()

                if self.unseen_level_handling == "Median":
                    self.unseen_levels_encoding_dict[c] = X_temp[c].median()

                if self.unseen_level_handling == "Lowest":
                    self.unseen_levels_encoding_dict[c] = X_temp[c].min()

                if self.unseen_level_handling == "Highest":
                    self.unseen_levels_encoding_dict[c] = X_temp[c].max()

        return self

    def map_imputation_values(self, X: pd.DataFrame) -> pd.DataFrame:
        """maps columns defined by self.columns in X according the the corresponding mapping dictionary contained in self.mappings

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform.

        Returns
        -------
        X : pd.DataFrame
            input dataframe with mappings applied
        """
        for c in self.columns:
            X[c] = X[c].map(self.mappings[c]).astype(self.return_type)

        return X

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform method to apply mean response encoding stored in the mappings attribute to
        each column in the columns attribute.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        N.B. In the mutli-level case, this method briefly overwrites the self.columns attribute, but sets
        it back to the original value at the end.

        Parameters
        ----------
        X : pd.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """
        X = super().transform(X)

        if self.level:
            for response_level in self.response_levels:
                for column in self.columns:
                    X[column + "_" + response_level] = X[column]
            temp_columns = self.columns
            # Temporarily overwriting self.columns to use BaseMappingTransformMixin
            self.columns = self.mapped_columns

        if self.unseen_level_handling:
            self.check_is_fitted(["mappings"])
            unseen_indices = {}
            for c in self.columns:
                # finding rows with values not in the keys of mappings dictionary
                unseen_indices[c] = X[~X[c].isin(self.mappings[c].keys())].index
            X = self.map_imputation_values(X)
            for c in self.columns:
                X.loc[unseen_indices[c], c] = self.unseen_levels_encoding_dict[c]
        else:
            self.check_mappable_rows(X)
            X = self.map_imputation_values(X)

        if self.level:
            # Setting self.columns back so that the transformer object is unchanged after transform is called
            self.columns = temp_columns
            X = X.drop(columns=self.columns)

        return X


class OrdinalEncoderTransformer(BaseNominalTransformer, BaseMappingTransformMixin):
    """Transformer to encode categorical variables into ascending rank-ordered integer values variables by mapping
    it's levels to the target-mean response for that level.
    Values will be sorted in ascending order only i.e. categorical level with lowest target mean response to
    be encoded as 1, the next highest value as 2 and so on.

    If a categorical variable contains null values these will not be transformed.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    weights_column : str or None
        Weights column to use when calculating the mean response.

    mappings : dict
        Created in fit. Dict of key (column names) value (mapping of categorical levels to numeric,
        ordinal encoded response values) pairs.

    """

    def __init__(
        self,
        columns: str | list[str] | None = None,
        weights_column: str | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if weights_column is not None and type(weights_column) is not str:
            msg = f"{self.classname()}: weights_column should be a str"
            raise TypeError(msg)

        self.weights_column = weights_column

        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

    def fit(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """Identify mapping of categorical levels to rank-ordered integer values by target-mean in ascending order.

        If the user specified the weights_column arg in when initialising the transformer
        the weighted mean response will be calculated using that column.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform and response_column column
            specified when object was initialised.

        y : pd.Series
            Response column or target.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}

        if (
            self.weights_column is not None
            and self.weights_column not in X.columns.to_numpy()
        ):
            msg = f"{self.classname()}: weights column {self.weights_column} not in X"
            raise ValueError(msg)

        response_null_count = y.isna().sum()

        if response_null_count > 0:
            msg = f"{self.classname()}: y has {response_null_count} null values"
            raise ValueError(msg)

        X_y = self._combine_X_y(X, y)
        response_column = "_temporary_response"

        for c in self.columns:
            if self.weights_column is None:
                # get the indexes of the sorted target mean-encoded dict
                _idx_target_mean = list(
                    X_y.groupby([c])[response_column]
                    .mean()
                    .sort_values(ascending=True, kind="mergesort")
                    .index,
                )

                # create a dictionary whose keys are the levels of the categorical variable
                # sorted ascending by their target-mean value
                # and whose values are ascending ordinal integers
                ordinal_encoded_dict = {
                    k: _idx_target_mean.index(k) + 1 for k in _idx_target_mean
                }

                self.mappings[c] = ordinal_encoded_dict

            else:
                groupby_sum = X_y.groupby([c])[
                    [response_column, self.weights_column]
                ].sum()

                # get the indexes of the sorted target mean-encoded dict
                _idx_target_mean = list(
                    (groupby_sum[response_column] / groupby_sum[self.weights_column])
                    .sort_values(ascending=True, kind="mergesort")
                    .index,
                )

                # create a dictionary whose keys are the levels of the categorical variable
                # sorted ascending by their target-mean value
                # and whose values are ascending ordinal integers
                ordinal_encoded_dict = {
                    k: _idx_target_mean.index(k) + 1 for k in _idx_target_mean
                }

                self.mappings[c] = ordinal_encoded_dict

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform method to apply ordinal encoding stored in the mappings attribute to
        each column in the columns attribute. This maps categorical levels to rank-ordered integer values by target-mean in ascending order.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed data with levels mapped to ordinal encoded values for categorical variables.

        """
        super(BaseNominalTransformer, self).columns_check(X)

        self.check_mappable_rows(X)

        return BaseMappingTransformMixin.transform(self, X)


class OneHotEncodingTransformer(BaseNominalTransformer, OneHotEncoder):
    """Transformer to convert cetegorical variables into dummy columns.

    Extends the sklearn OneHotEncoder class to provide easy renaming of dummy columns.

    Parameters
    ----------
    columns : str or list of strings or None, default = None
        Names of columns to transform. If the default of None is supplied all object and category
        columns in X are used.

    separator : str
        Used to create dummy column names, the name will take
        the format [categorical feature][separator][category level]

    drop_original : bool, default = False
        Should original columns be dropped after creating dummy fields?

    copy : bool, default = True
        Should X be copied prior to transform?

    verbose : bool, default = True
        Should warnings/checkmarks get displayed?

    **kwargs
        Arbitrary keyword arguments passed onto sklearn OneHotEncoder.init method.

    Attributes
    ----------
    separator : str
        Separator used in naming for dummy columns.

    drop_original : bool
        Should original columns be dropped after creating dummy fields?

    """

    def __init__(
        self,
        columns: str | list[str] | None = None,
        separator: str = "_",
        drop_original: bool = False,
        copy: bool = True,
        verbose: bool = False,
        dtype: np.int8 = np.int8,
        **kwargs: dict[str, bool],
    ) -> None:
        BaseNominalTransformer.__init__(
            self,
            columns=columns,
            copy=copy,
            verbose=verbose,
        )

        # Set attributes for scikit-learn'S OneHotEncoder
        OneHotEncoder.__init__(
            self,
            sparse=False,
            handle_unknown="ignore",
            dtype=dtype,
            **kwargs,
        )

        # Set other class attrributes
        self.separator = separator
        self.drop_original = drop_original

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Gets list of levels for each column to be transformed. This defines which dummy columns
        will be created in transform.

        Parameters
        ----------
        X : pd.DataFrame
            Data to identify levels from.

        y : None
            Ignored. This parameter exists only for compatibility with sklearn.pipeline.Pipeline.

        """
        BaseNominalTransformer.fit(self, X=X, y=y)

        # Check for nulls
        for c in self.columns:
            if X[c].isna().sum() > 0:
                raise ValueError(
                    f"{self.classname()}: column %s has nulls - replace before proceeding"
                    % c,
                )

        # Check each field has less than 100 categories/levels
        for c in self.columns:
            levels = X[c].unique().tolist()

            if len(levels) > 100:
                raise ValueError(
                    f"{self.classname()}: column %s has over 100 unique values - consider another type of encoding"
                    % c,
                )

        OneHotEncoder.fit(self, X=X[self.columns], y=y)

        return self

    def _get_feature_names(
        self,
        input_features: list[str],
        **kwargs: dict[str, bool],
    ) -> list[str]:
        """Function to access the get_feature_names attribute of the scikit learn attribute. Will return the output columns of the OHE transformer.

        In scikit learn 1.0 "get_feature_names" was deprecated and then replaced with "get_feature_names_out" in version 1.2. The logic in this
        function will call the correct attribute, or raise an error if it can't be found.

        Parameters
        ----------
        input_features : list(str)
            Input columns being transformed by the OHE transformer.

        kwargs : dict
            Keyword arguments to be passed on to the scikit learn attriute.
        """
        if hasattr(self, "get_feature_names"):
            input_columns = self.get_feature_names(
                input_features=input_features,
                **kwargs,
            )

        elif hasattr(self, "get_feature_names_out"):
            input_columns = self.get_feature_names_out(
                input_features=input_features,
                **kwargs,
            )

        else:
            msg = "Cannot access scikit learn OneHotEncoder get_feature_names method, may be a version issue"
            raise AttributeError(msg)

        return input_columns

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Create new dummy columns from categorical fields.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply one hot encoding to.

        Returns
        -------
        X_transformed : pd.DataFrame
            Transformed input X with dummy columns derived from categorical columns added. If drop_original
            = True then the original categorical columns that the dummies are created from will not be in
            the output X.

        """
        self.check_is_fitted(["separator"])
        self.check_is_fitted(["drop_original"])

        self.columns_check(X)

        # Check for nulls
        for c in self.columns:
            if X[c].isna().sum() > 0:
                raise ValueError(
                    f"{self.classname()}: column %s has nulls - replace before proceeding"
                    % c,
                )

        X = BaseNominalTransformer.transform(self, X)

        # Apply OHE transform
        X_transformed = OneHotEncoder.transform(self, X[self.columns])

        input_columns = self._get_feature_names(input_features=self.columns)

        X_transformed = pd.DataFrame(
            X_transformed,
            columns=input_columns,
            index=X.index,
        )

        # Rename dummy fields if separator is specified
        if self.separator != "_":
            old_names = [
                c + "_" + str(lvl)
                for i, c in enumerate(self.columns)
                for lvl in self.categories_[i]
            ]
            new_names = [
                c + self.separator + str(lvl)
                for i, c in enumerate(self.columns)
                for lvl in self.categories_[i]
            ]

            X_transformed = X_transformed.rename(
                columns=dict(zip(old_names, new_names)),
            )

        # Print warning for unseen levels
        if self.verbose:
            for i, c in enumerate(self.columns):
                unseen_levels = set(X[c].unique().tolist()) - set(self.categories_[i])

                if len(unseen_levels) > 0:
                    warnings.warn(
                        f"{self.classname()}: column %s has unseen categories: %s"
                        % (c, unseen_levels),
                        stacklevel=2,
                    )

        # Drop original columns
        if self.drop_original:
            X = X.drop(self.columns, axis=1)

        # Concatenate original and new dummy fields
        return pd.concat((X, X_transformed), axis=1)



"""This module contains transformers that apply string functions."""

from __future__ import annotations

import pandas as pd

from tubular.base import BaseTransformer


class SeriesStrMethodTransformer(BaseTransformer):
    """Tranformer that applies a pandas.Series.str method.

    Transformer assigns the output of the method to a new column. It is possible to
    supply other key word arguments to the transform method, which will be passed to the
    pandas.Series.str method being called.

    Be aware it is possible to supply incompatible arguments to init that will only be
    identified when transform is run. This is because there are many combinations of method, input
    and output sizes. Additionally some methods may only work as expected when called in
    transform with specific key word arguments.

    Parameters
    ----------
    new_column_name : str
        The name of the column to be assigned to the output of running the pd.Series.str in transform.

    pd_method_name : str
        The name of the pandas.Series.str method to call.

    columns : str
        Column to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object. Note this has no default value so
        the user has to specify the columns when initialising the transformer. This is avoid likely
        when the user forget to set columns, in this case all columns would be picked up when super
        transform runs.

    pd_method_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.Series.str method when it is called.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    new_column_name : str
        The name of the column or columns to be assigned to the output of running the
        pd.Series.str in transform.

    pd_method_name : str
        The name of the pd.Series.str method to call.

    """

    def __init__(
        self,
        new_column_name: str,
        pd_method_name: str,
        columns: str,
        pd_method_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if type(columns) is list and len(columns) > 1:
            msg = f"{self.classname()}: columns arg should contain only 1 column name but got {len(columns)}"
            raise ValueError(msg)

        super().__init__(columns=columns, **kwargs)

        if type(new_column_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(new_column_name)}) for new_column_name, must be str"
            raise TypeError(msg)

        if type(pd_method_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(pd_method_name)}) for pd_method_name, expecting str"
            raise TypeError(msg)

        if pd_method_kwargs is None:
            pd_method_kwargs = {}
        else:
            if type(pd_method_kwargs) is not dict:
                msg = f"{self.classname()}: pd_method_kwargs should be a dict but got type {type(pd_method_kwargs)}"
                raise TypeError(msg)

        for i, k in enumerate(pd_method_kwargs.keys()):
            if type(k) is not str:
                msg = f"{self.classname()}: unexpected type ({type(k)}) for pd_method_kwargs key in position {i}, must be str"
                raise TypeError(msg)

        self.new_column_name = new_column_name
        self.pd_method_name = pd_method_name
        self.pd_method_kwargs = pd_method_kwargs

        try:
            ser = pd.Series(["a"])
            getattr(ser.str, pd_method_name)

        except Exception as err:
            msg = f'{self.classname()}: error accessing "str.{pd_method_name}" method on pd.Series object - pd_method_name should be a pd.Series.str method'
            raise AttributeError(msg) from err

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform specific column on input pandas.DataFrame (X) using the given pandas.Series.str method and
        assign the output back to column in X.

        Any keyword arguments set in the pd_method_kwargs attribute are passed onto the pd.Series.str method
        when calling it.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column (self.new_column_name) added. These contain the output of
            running the pd.Series.str method.

        """
        X = super().transform(X)

        X[self.new_column_name] = getattr(X[self.columns[0]].str, self.pd_method_name)(
            **self.pd_method_kwargs,
        )

        return X


class StringConcatenator(BaseTransformer):
    """Transformer to combine data from specified columns, of mixed datatypes, into a new column containing one string.

    Parameters
    ----------
    columns : str or list of str
        Columns to concatenate.
    new_column : str, default = "new_column"
        New column name
    separator : str, default = " "
        Separator for the new string value
    """

    def __init__(
        self,
        columns: str | list[str],
        new_column: str = "new_column",
        separator: str = " ",
    ) -> None:
        super().__init__(columns=columns, copy=True)

        if not isinstance(new_column, str):
            msg = f"{self.classname()}: new_column should be a str"
            raise TypeError(msg)

        self.new_column = new_column

        if not isinstance(separator, str):
            msg = f"{self.classname()}: The separator should be a str"
            raise TypeError(msg)

        self.separator = separator

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Combine data from specified columns, of mixed datatypes, into a new column containing one string.

        Parameters
        ----------
        X : df
            Data to concatenate values on.

        Returns
        -------
        X : df
            Returns a dataframe with concatenated values.

        """
        X = super().transform(X)

        X[self.new_column] = (
            X[self.columns].astype(str).apply(lambda x: self.separator.join(x), axis=1)
        )

        return X



"""This module contains a transformer that applies capping to numeric columns."""

from __future__ import annotations

import datetime
import itertools
import warnings

import numpy as np
import pandas as pd

from tubular.base import BaseTransformer


class BaseDateTransformer(BaseTransformer):
    """
    Transformer with date data checks needed by all transformers taking dates as input data.
    """

    def _generate_is_datetime64_dict(self, X: pd.DataFrame) -> None:
        """
        Function to generate a dictionary attribute storing the result of pd.api.types.is_datetime64_any_dtype()
        for each column in self.columns to avoid repeated calls to this function.
        """

        self._is_datetime64_dict = {}

        for col in self.columns:
            self._is_datetime64_dict[col] = pd.api.types.is_datetime64_any_dtype(X[col])

    def check_columns_are_date_or_datetime(self, X: pd.DataFrame) -> None:
        "Raise a type error if a column to be operated on is not a datetime.datetime or datetime.date object"

        self._generate_is_datetime64_dict(X)

        for col in self.columns:
            if (
                not self._is_datetime64_dict[col]
                and pd.api.types.infer_dtype(X[col]) != "date"
            ):
                msg = f"{self.classname()}: {col} should be datetime64 or date type but got {X[col].dtype}"
                raise TypeError(msg)

    def cast_columns_to_datetime(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Check whether columns are datetime64 or date type. If not, an error will be raised. If one is datetime.date
        it will be cast to datetime64.
        """

        self.check_columns_are_date_or_datetime(X)

        temp = X[self.columns].copy()

        for column in self.columns:
            if not self._is_datetime64_dict[column]:
                temp[column] = pd.to_datetime(X[column])

                warnings.warn(
                    f"""
                    {self.classname()}: temporarily cast {column} from datetime64 to date before transforming in order to apply the datetime method.

                    This will artificially increase the precision of each data point in the column. Original column not changed.
                    """,
                    stacklevel=2,
                )

        return temp

    def _cast_non_matching_columns(
        self,
        temp: pd.DataFrame,
        column_A_name: str,
        column_B_name: str,
    ) -> pd.DataFrame:
        """
        Helper function that asymetrically compares column A to column B and casts column B to match column A. This will need calling twice.

        If any casting is done a user warning is raised.

        Operation is:
        - check column A is not datetime 64 (assumed that only other option is 'date' type based on prior checks)
        - check column B is datetime 64
        - if both of the above are true, cast column B to 'date'
        """

        if (
            not self._is_datetime64_dict[column_A_name]
            and self._is_datetime64_dict[column_B_name]
        ):
            temp[column_B_name] = temp[column_B_name].apply(lambda x: x.date())

            warnings.warn(
                f"""
                {self.classname()}: temporarily cast {column_B_name} from datetime64 to date before transforming in order to match {column_A_name}.

                Some precision may be lost from {column_B_name}. Original column not changed.
                """,
                stacklevel=2,
            )

        return temp

    def match_column_dtypes(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Check the dtype of the two columns to be compared by the transformer. If one is datetime.date and one is
        datetime.datetime, the datetime.datetime column will be cast to datetime.date.

        The casting is done this way to avoid artificially creating precision we don't have by going from a date to a datetime object
        """
        if len(self.columns) != 2:
            msg = f"{self.classname}: match_column_dtypes() called expecting 2 columns but got {len(self.columns)}"
            raise ValueError(msg)

        self.check_columns_are_date_or_datetime(X)

        temp = X[self.columns].copy()

        for column_one_name, column_two_name in itertools.permutations(self.columns):

            temp = self._cast_non_matching_columns(
                temp,
                column_one_name,
                column_two_name,
            )

        return temp


class DateDiffLeapYearTransformer(BaseDateTransformer):
    """Transformer to calculate the number of years between two dates.

    Parameters
    ----------
    column_lower : str
        Name of date column to subtract.

    column_upper : str
        Name of date column to subtract from.

    new_column_name : str
        Name for the new year column.

    drop_cols : bool
        Flag for whether to drop the original columns.

    missing_replacement : int/float/str
        Value to output if either the lower date value or the upper date value are
        missing. Default value is None.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    column_lower : str
        Name of date column to subtract. This attribute is not for use in any method,
        use 'columns' instead. Here only as a fix to allow string representation of transformer.

    column_upper : str
        Name of date column to subtract from. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : list
        List containing column names for transformation in format [column_lower, column_upper]

    new_column_name : str, default = None
        Name given to calculated datediff column. If None then {column_upper}_{column_lower}_datediff
        will be used.

    drop_cols : bool
        Indicator whether to drop old columns during transform method.

    """

    def __init__(
        self,
        column_lower: str,
        column_upper: str,
        drop_cols: bool,
        new_column_name: str | None = None,
        missing_replacement: int | float | str | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if not isinstance(column_lower, str):
            msg = f"{self.classname()}: column_lower should be a str"
            raise TypeError(msg)

        if not isinstance(column_upper, str):
            msg = f"{self.classname()}: column_upper should be a str"
            raise TypeError(msg)

        if new_column_name is not None:
            if not isinstance(new_column_name, str):
                msg = f"{self.classname()}: new_column_name should be a str"
                raise TypeError(msg)

            self.new_column_name = new_column_name

        else:
            self.new_column_name = f"{column_upper}_{column_lower}_datediff"

        if not isinstance(drop_cols, bool):
            msg = f"{self.classname()}: drop_cols should be a bool"
            raise TypeError(msg)

        if (missing_replacement) and (
            type(missing_replacement) not in [int, float, str]
        ):
            msg = f"{self.classname()}: if not None, missing_replacement should be an int, float or string"
            raise TypeError(msg)

        super().__init__(columns=[column_lower, column_upper], **kwargs)

        self.drop_cols = drop_cols
        self.missing_replacement = missing_replacement

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper

    def calculate_age(self, row: pd.Series) -> int:
        """Function to calculate age from two date columns in a pd.DataFrame.

        This function, although slower than the np.timedelta64 solution (or something
        similar), accounts for leap years to accurately calculate age for all values.

        Parameters
        ----------
        row : pd.Series
            Named pandas series, with lower_date_name and upper_date_name as index values.

        Returns
        -------
        age : int
            Year gap between the upper and lower date values passes.

        """
        if not isinstance(row, pd.Series):
            msg = f"{self.classname()}: row should be a pd.Series"
            raise TypeError(msg)

        if (pd.isna(row[self.columns[0]])) or (pd.isna(row[self.columns[1]])):
            return self.missing_replacement

        age = row[self.columns[1]].year - row[self.columns[0]].year

        if age > 0:
            if (row[self.columns[1]].month, row[self.columns[1]].day) < (
                row[self.columns[0]].month,
                row[self.columns[0]].day,
            ):
                age += -1
        elif age < 0 and (row[self.columns[1]].month, row[self.columns[1]].day) > (
            row[self.columns[0]].month,
            row[self.columns[0]].day,
        ):
            age += 1

        return age

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Calculate year gap between the two provided columns.

        New column is created under the 'new_column_name', and optionally removes the
        old date columns.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing column_upper and column_lower.

        Returns
        -------
        X : pd.DataFrame
            Transformed data with new_column_name column.

        """

        self.check_columns_are_date_or_datetime(X)

        X = super().transform(X)

        X[self.new_column_name] = X.apply(lambda x: self.calculate_age(x), axis=1)

        if self.drop_cols:
            X = X.drop(self.columns, axis=1)

        return X


class DateDifferenceTransformer(BaseDateTransformer):
    """Class to transform calculate the difference between 2 date fields in specified units.

    Parameters
    ----------
    column_lower : str
        Name of first date column, difference will be calculated as column_upper - column_lower
    column_upper : str
        Name of second date column, difference will be calculated as column_upper - column_lower
    new_column_name : str, default = None
        Name given to calculated datediff column. If None then {column_upper}_{column_lower}_datediff_{units}
        will be used.
    units : str, default = 'D'
        Numpy datetime units, accepted values are 'D', 'h', 'm', 's'
    copy : bool, default = True
        Should X be copied prior to transform?
    verbose: bool, default = False
    """

    def __init__(
        self,
        column_lower: str,
        column_upper: str,
        new_column_name: str | None = None,
        units: str = "D",
        copy: bool = True,
        verbose: bool = False,
    ) -> None:
        if type(column_lower) is not str:
            msg = f"{self.classname()}: column_lower must be a str"
            raise TypeError(msg)

        if type(column_upper) is not str:
            msg = f"{self.classname()}: column_upper must be a str"
            raise TypeError(msg)

        columns = [column_lower, column_upper]

        accepted_values_units = [
            "D",
            "h",
            "m",
            "s",
        ]

        if type(units) is not str:
            msg = f"{self.classname()}: units must be a str"
            raise TypeError(msg)

        if units not in accepted_values_units:
            msg = f"{self.classname()}: units must be one of {accepted_values_units}, got {units}"
            raise ValueError(msg)

        self.units = units

        if new_column_name is not None:
            if type(new_column_name) is not str:
                msg = f"{self.classname()}: new_column_name must be a str"
                raise TypeError(msg)

            self.new_column_name = new_column_name

        else:
            self.new_column_name = f"{column_upper}_{column_lower}_datediff_{units}"

        super().__init__(columns=columns, copy=copy, verbose=verbose)

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Calculate the difference between the given fields in the specified units.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        """

        X = super().transform(X)

        temp = self.match_column_dtypes(X)

        X[self.new_column_name] = (
            temp[self.columns[1]] - temp[self.columns[0]]
        ) / np.timedelta64(1, self.units)

        return X


class ToDatetimeTransformer(BaseTransformer):
    """Class to transform convert specified column to datetime.

    Class simply uses the pd.to_datetime method on the specified column.

    Parameters
    ----------
    column : str
        Name of the column to convert to datetime.

    new_column_name : str
        Name given to the new datetime column.

    to_datetime_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.to_datetime method when it is called in transform.

    **kwargs
        Arbitrary keyword arguments passed onto pd.to_datetime().

    """

    def __init__(
        self,
        column: str,
        new_column_name: str,
        to_datetime_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if type(column) is not str:
            msg = f"{self.classname()}: column should be a single str giving the column to transform to datetime"
            raise TypeError(msg)

        if type(new_column_name) is not str:
            msg = f"{self.classname()}: new_column_name must be a str"
            raise TypeError(msg)

        if to_datetime_kwargs is None:
            to_datetime_kwargs = {}
        else:
            if type(to_datetime_kwargs) is not dict:
                msg = f"{self.classname()}: to_datetime_kwargs should be a dict but got type {type(to_datetime_kwargs)}"
                raise TypeError(msg)

            for i, k in enumerate(to_datetime_kwargs.keys()):
                if type(k) is not str:
                    msg = f"{self.classname()}: unexpected type ({type(k)}) for to_datetime_kwargs key in position {i}, must be str"
                    raise TypeError(msg)

        self.to_datetime_kwargs = to_datetime_kwargs
        self.new_column_name = new_column_name

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

        super().__init__(columns=[column], **kwargs)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Convert specified column to datetime using pd.to_datetime.

        Parameters
        ----------
        X : pd.DataFrame
            Data with column to transform.

        """
        X = super().transform(X)

        X[self.new_column_name] = pd.to_datetime(
            X[self.columns[0]],
            **self.to_datetime_kwargs,
        )

        return X


class SeriesDtMethodTransformer(BaseDateTransformer):
    """Tranformer that applies a pandas.Series.dt method.

    Transformer assigns the output of the method to a new column. It is possible to
    supply other key word arguments to the transform method, which will be passed to the
    pandas.Series.dt method being called.

    Be aware it is possible to supply incompatible arguments to init that will only be
    identified when transform is run. This is because there are many combinations of method, input
    and output sizes. Additionally some methods may only work as expected when called in
    transform with specific key word arguments.

    Parameters
    ----------
    new_column_name : str
        The name of the column to be assigned to the output of running the pandas method in transform.

    pd_method_name : str
        The name of the pandas.Series.dt method to call.

    column : str
        Column to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object. Note this has no default value so
        the user has to specify the columns when initialising the transformer. This is avoid likely
        when the user forget to set columns, in this case all columns would be picked up when super
        transform runs.

    pd_method_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.Series.dt method when it is called.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    column : str
        Name of column to apply transformer to. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : str
        Column name for transformation.

    new_column_name : str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    pd_method_kwargs : dict
        Dictionary of keyword arguments to call the pd.Series.dt method with.

    """

    def __init__(
        self,
        new_column_name: str,
        pd_method_name: str,
        column: str,
        pd_method_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if type(column) is not str:
            msg = f"{self.classname()}: column should be a str but got {type(column)}"
            raise TypeError(msg)

        super().__init__(columns=column, **kwargs)

        if type(new_column_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(new_column_name)}) for new_column_name, must be str"
            raise TypeError(msg)

        if type(pd_method_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(pd_method_name)}) for pd_method_name, expecting str"
            raise TypeError(msg)

        if pd_method_kwargs is None:
            pd_method_kwargs = {}
        else:
            if type(pd_method_kwargs) is not dict:
                msg = f"{self.classname()}: pd_method_kwargs should be a dict but got type {type(pd_method_kwargs)}"
                raise TypeError(msg)

            for i, k in enumerate(pd_method_kwargs.keys()):
                if type(k) is not str:
                    msg = f"{self.classname()}: unexpected type ({type(k)}) for pd_method_kwargs key in position {i}, must be str"
                    raise TypeError(msg)

        self.new_column_name = new_column_name
        self.pd_method_name = pd_method_name
        self.pd_method_kwargs = pd_method_kwargs

        try:
            ser = pd.Series(
                [datetime.datetime(2020, 12, 21, tzinfo=datetime.timezone.utc)],
            )
            getattr(ser.dt, pd_method_name)

        except Exception as err:
            msg = f'{self.classname()}: error accessing "dt.{pd_method_name}" method on pd.Series object - pd_method_name should be a pd.Series.dt method'
            raise AttributeError(msg) from err

        if callable(getattr(ser.dt, pd_method_name)):
            self._callable = True

        else:
            self._callable = False

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform specific column on input pandas.DataFrame (X) using the given pandas.Series.dt method and
        assign the output back to column in X.

        Any keyword arguments set in the pd_method_kwargs attribute are passed onto the pd.Series.dt method
        when calling it.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column (self.new_column_name) added. These contain the output of
            running the pd.Series.dt method.

        """
        X = super().transform(X)

        temp = self.cast_columns_to_datetime(X)

        if self._callable:
            X[self.new_column_name] = getattr(
                temp[self.columns[0]].dt,
                self.pd_method_name,
            )(**self.pd_method_kwargs)

        else:
            X[self.new_column_name] = getattr(
                temp[self.columns[0]].dt,
                self.pd_method_name,
            )

        return X


class BetweenDatesTransformer(BaseDateTransformer):
    """Transformer to generate a boolean column indicating if one date is between two others.

    If not all column_lower values are less than or equal to column_upper when transform is run
    then a warning will be raised.

    Parameters
    ----------
    column_lower : str
        Name of column containing the lower date range values.

    column_between : str
        Name of column to check if it's values fall between column_lower and column_upper.

    column_upper : str
        Name of column containing the upper date range values.

    new_column_name : str
        Name for new column to be added to X.

    lower_inclusive : bool, defualt = True
        If lower_inclusive is True the comparison to column_lower will be column_lower <=
        column_between, otherwise the comparison will be column_lower < column_between.

    upper_inclusive : bool, defualt = True
        If upper_inclusive is True the comparison to column_upper will be column_between <=
        column_upper, otherwise the comparison will be column_between < column_upper.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    column_lower : str
        Name of date column to subtract. This attribute is not for use in any method,
        use 'columns' instead. Here only as a fix to allow string representation of transformer.

    column_upper : str
        Name of date column to subtract from. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    column_between : str
        Name of column to check if it's values fall between column_lower and column_upper. This attribute
        is not for use in any method, use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : list
        Contains the names of the columns to compare in the order [column_lower, column_between
        column_upper].

    new_column_name : str
        new_column_name argument passed when initialising the transformer.

    lower_inclusive : bool
        lower_inclusive argument passed when initialising the transformer.

    upper_inclusive : bool
        upper_inclusive argument passed when initialising the transformer.

    """

    def __init__(
        self,
        column_lower: str,
        column_between: str,
        column_upper: str,
        new_column_name: str,
        lower_inclusive: bool = True,
        upper_inclusive: bool = True,
        **kwargs: dict[str, bool],
    ) -> None:
        if type(column_lower) is not str:
            msg = f"{self.classname()}: column_lower should be str"
            raise TypeError(msg)

        if type(column_between) is not str:
            msg = f"{self.classname()}: column_between should be str"
            raise TypeError(msg)

        if type(column_upper) is not str:
            msg = f"{self.classname()}: column_upper should be str"
            raise TypeError(msg)

        if type(new_column_name) is not str:
            msg = f"{self.classname()}: new_column_name should be str"
            raise TypeError(msg)

        if type(lower_inclusive) is not bool:
            msg = f"{self.classname()}: lower_inclusive should be a bool"
            raise TypeError(msg)

        if type(upper_inclusive) is not bool:
            msg = f"{self.classname()}: upper_inclusive should be a bool"
            raise TypeError(msg)

        self.new_column_name = new_column_name
        self.lower_inclusive = lower_inclusive
        self.upper_inclusive = upper_inclusive

        super().__init__(columns=[column_lower, column_between, column_upper], **kwargs)

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper
        self.column_between = column_between

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform - creates column indicating if middle date is between the other two.

        If not all column_lower values are less than or equal to column_upper when transform is run
        then a warning will be raised.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column (self.new_column_name) added. This column is
            boolean and indicates if the middle column is between the other 2.

        """
        X = super().transform(X)

        self.check_columns_are_date_or_datetime(X)

        if not (X[self.columns[0]] <= X[self.columns[2]]).all():
            warnings.warn(
                f"{self.classname()}: not all {self.columns[2]} are greater than or equal to {self.columns[0]}",
                stacklevel=2,
            )

        if self.lower_inclusive:
            lower_comparison = X[self.columns[0]] <= X[self.columns[1]]

        else:
            lower_comparison = X[self.columns[0]] < X[self.columns[1]]

        if self.upper_inclusive:
            upper_comparison = X[self.columns[1]] <= X[self.columns[2]]

        else:
            upper_comparison = X[self.columns[1]] < X[self.columns[2]]

        X[self.new_column_name] = lower_comparison & upper_comparison

        return X


class DatetimeInfoExtractor(BaseDateTransformer):
    """Transformer to extract various features from datetime var.

    Parameters
    ----------
    columns : str or list
        datetime columns to extract information from

    include : list of str, default = ["timeofday", "timeofmonth", "timeofyear", "dayofweek"]
        Which datetime categorical information to extract

    datetime_mappings : dict, default = {}
        Optional argument to define custom mappings for datetime values.
        Keys of the dictionary must be contained in `include`
        All possible values of each feature must be included in the mappings,
        ie, a mapping for `dayofweek` must include all values 0-6;
        datetime_mappings = {"dayofweek": {"week": [0, 1, 2, 3, 4],
                                           "weekend": [5, 6]}}
        The values for the mapping array must be iterable;
        datetime_mappings = {"timeofday": {"am": range(0, 12),
                                           "pm": range(12, 24)}}
        The required ranges for each mapping are:
            timeofday: 0-23
            timeofmonth: 1-31
            timeofyear: 1-12
            dayofweek: 0-6

        If in include but no mappings provided default values will be used as follows:
           timeofday_mapping = {
                "night": range(0, 6),  # Midnight - 6am
                "morning": range(6, 12),  # 6am - Noon
                "afternoon": range(12, 18),  # Noon - 6pm
                "evening": range(18, 24),  # 6pm - Midnight
            }
            timeofmonth_mapping = {
                "start": range(0, 11),
                "middle": range(11, 21),
                "end": range(21, 32),
            }
            timeofyear_mapping = {
                "spring": range(3, 6),  # Mar, Apr, May
                "summer": range(6, 9),  # Jun, Jul, Aug
                "autumn": range(9, 12),  # Sep, Oct, Nov
                "winter": [12, 1, 2],  # Dec, Jan, Feb
            }
            dayofweek_mapping = {
                "monday": [0],
                "tuesday": [1],
                "wednesday": [2],
                "thursday": [3],
                "friday": [4],
                "saturday": [5],
                "sunday": [6],
            }


    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    include : list of str, default = ["timeofday", "timeofmonth", "timeofyear", "dayofweek"]
        Which datetime categorical information to extract

    datetime_mappings : dict, default = None
        Optional argument to define custom mappings for datetime values.

    """

    def __init__(
        self,
        columns: str | list[str],
        include: str | list[str] | None = None,
        datetime_mappings: dict[
            str,
        ]
        | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if include is None:
            include = ["timeofday", "timeofmonth", "timeofyear", "dayofweek"]
        else:
            if type(include) is not list:
                msg = f"{self.classname()}: include should be List"
                raise TypeError(msg)

        if datetime_mappings is None:
            datetime_mappings = {}
        else:
            if type(datetime_mappings) is not dict:
                msg = f"{self.classname()}: datetime_mappings should be Dict"
                raise TypeError(msg)

        super().__init__(columns=columns, **kwargs)

        for var in include:
            if var not in [
                "timeofday",
                "timeofmonth",
                "timeofyear",
                "dayofweek",
            ]:
                msg = f'{self.classname()}: elements in include should be in ["timeofday", "timeofmonth", "timeofyear", "dayofweek"]'
                raise ValueError(msg)

        if datetime_mappings != {}:
            for key, mapping in datetime_mappings.items():
                if type(mapping) is not dict:
                    msg = f"{self.classname()}: values in datetime_mappings should be dict"
                    raise TypeError(msg)
                if key not in include:
                    msg = f"{self.classname()}: keys in datetime_mappings should be in include"
                    raise ValueError(msg)

        self.include = include
        self.datetime_mappings = datetime_mappings
        self.mappings_provided = self.datetime_mappings.keys()

        # Select correct mapping either from default or user input

        if ("timeofday" in include) and ("timeofday" in self.mappings_provided):
            timeofday_mapping = self.datetime_mappings["timeofday"]
        elif "timeofday" in include:  # Choose default mapping
            timeofday_mapping = {
                "night": range(0, 6),  # Midnight - 6am
                "morning": range(6, 12),  # 6am - Noon
                "afternoon": range(12, 18),  # Noon - 6pm
                "evening": range(18, 24),  # 6pm - Midnight
            }

        if ("timeofmonth" in include) and ("timeofmonth" in self.mappings_provided):
            timeofmonth_mapping = self.datetime_mappings["timeofmonth"]
        elif "timeofmonth" in include:  # Choose default mapping
            timeofmonth_mapping = {
                "start": range(0, 11),
                "middle": range(11, 21),
                "end": range(21, 32),
            }

        if ("timeofyear" in include) and ("timeofyear" in self.mappings_provided):
            timeofyear_mapping = self.datetime_mappings["timeofyear"]
        elif "timeofyear" in include:  # Choose default mapping
            timeofyear_mapping = {
                "spring": range(3, 6),  # Mar, Apr, May
                "summer": range(6, 9),  # Jun, Jul, Aug
                "autumn": range(9, 12),  # Sep, Oct, Nov
                "winter": [12, 1, 2],  # Dec, Jan, Feb
            }

        if ("dayofweek" in include) and ("dayofweek" in self.mappings_provided):
            dayofweek_mapping = self.datetime_mappings["dayofweek"]
        elif "dayofweek" in include:  # Choose default mapping
            dayofweek_mapping = {
                "monday": [0],
                "tuesday": [1],
                "wednesday": [2],
                "thursday": [3],
                "friday": [4],
                "saturday": [5],
                "sunday": [6],
            }

        # Invert dictionaries for quicker lookup

        if "timeofday" in include:
            self.timeofday_mapping = {
                vi: k for k, v in timeofday_mapping.items() for vi in v
            }
            if set(self.timeofday_mapping.keys()) != set(range(24)):
                msg = f"{self.classname()}: timeofday mapping dictionary should contain mapping for all hours between 0-23. {set(range(24)) - set(self.timeofday_mapping.keys())} are missing"
                raise ValueError(msg)

            # Check if all hours in dictionary
        else:
            self.timeofday_mapping = {}

        if "timeofmonth" in include:
            self.timeofmonth_mapping = {
                vi: k for k, v in timeofmonth_mapping.items() for vi in v
            }
            if set(self.timeofmonth_mapping.keys()) != set(range(32)):
                msg = f"{self.classname()}: timeofmonth mapping dictionary should contain mapping for all days between 1-31. {set(range(1, 32)) - set(self.timeofmonth_mapping.keys())} are missing"
                raise ValueError(msg)
        else:
            self.timeofmonth_mapping = {}

        if "timeofyear" in include:
            self.timeofyear_mapping = {
                vi: k for k, v in timeofyear_mapping.items() for vi in v
            }
            if set(self.timeofyear_mapping.keys()) != set(range(1, 13)):
                msg = f"{self.classname()}: timeofyear mapping dictionary should contain mapping for all months between 1-12. {set(range(1, 13)) - set(self.timeofyear_mapping.keys())} are missing"
                raise ValueError(msg)

        else:
            self.timeofyear_mapping = {}

        if "dayofweek" in include:
            self.dayofweek_mapping = {
                vi: k for k, v in dayofweek_mapping.items() for vi in v
            }
            if set(self.dayofweek_mapping.keys()) != set(range(7)):
                msg = f"{self.classname()}: dayofweek mapping dictionary should contain mapping for all days between 0-6. {set(range(7)) - set(self.dayofweek_mapping.keys())} are missing"
                raise ValueError(msg)

        else:
            self.dayofweek_mapping = {}

    def _map_values(self, value: int | float, interval: str) -> str:
        """Method to apply mappings for a specified interval ("timeofday", "timeofmonth", "timeofyear" or "dayofweek")
        from corresponding mapping attribute to a single value.

        Parameters
        ----------
        interval : str
            the time period to map "timeofday", "timeofmonth", "timeofyear" or "dayofweek"

        value : float or int
            the value to be mapped


        Returns
        -------
        str : str
            Mapped value
        """
        if type(value) is not float and type(value) is not int:
            msg = f"{self.classname()}: value should be float or int"
            raise TypeError(msg)

        errors = {
            "timeofday": "0-23",
            "dayofweek": "0-6",
            "timeofmonth": "1-31",
            "timeofyear": "1-12",
        }
        ranges = {
            "timeofday": (0, 24, 1),
            "dayofweek": (0, 7, 1),
            "timeofmonth": (1, 32, 1),
            "timeofyear": (1, 13, 1),
        }
        mappings = {
            "timeofday": self.timeofday_mapping,
            "dayofweek": self.dayofweek_mapping,
            "timeofmonth": self.timeofmonth_mapping,
            "timeofyear": self.timeofyear_mapping,
        }

        if (not np.isnan(value)) and (value not in np.arange(*ranges[interval])):
            msg = f"{self.classname()}: value for {interval} mapping  in self._map_values should be an integer value in {errors[interval]}"
            raise ValueError(msg)

        if np.isnan(value):
            return np.nan

        return mappings[interval][value]

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform - Extracts new features from datetime variables.

        Parameters
        ----------
        X : pd.DataFrame
            Data with columns to extract info from.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with added columns of extracted information.
        """
        X = super().transform(X)

        temp = self.cast_columns_to_datetime(X)

        for col in self.columns:
            if "timeofday" in self.include:
                X[col + "_timeofday"] = temp[col].dt.hour.apply(
                    self._map_values,
                    interval="timeofday",
                )

            if "timeofmonth" in self.include:
                X[col + "_timeofmonth"] = temp[col].dt.day.apply(
                    self._map_values,
                    interval="timeofmonth",
                )

            if "timeofyear" in self.include:
                X[col + "_timeofyear"] = temp[col].dt.month.apply(
                    self._map_values,
                    interval="timeofyear",
                )

            if "dayofweek" in self.include:
                X[col + "_dayofweek"] = temp[col].dt.weekday.apply(
                    self._map_values,
                    interval="dayofweek",
                )

        return X


class DatetimeSinusoidCalculator(BaseDateTransformer):
    """Transformer to derive a feature in a dataframe by calculating the
    sine or cosine of a datetime column in a given unit (e.g hour), with the option to scale
    period of the sine or cosine to match the natural period of the unit (e.g. 24).

    Parameters
    ----------
    columns : str or list
        Columns to take the sine or cosine of. Must be a datetime[64] column.

    method : str or list
        Argument to specify which function is to be calculated. Accepted values are 'sin', 'cos' or a list containing both.

    units : str or dict
        Which time unit the calculation is to be carried out on. Accepted values are 'year', 'month',
        'day', 'hour', 'minute', 'second', 'microsecond'.  Can be a string or a dict containing key-value pairs of column
        name and units to be used for that column.

    period : int, float or dict, default = 2*np.pi
        The period of the output in the units specified above. To leave the period of the sinusoid output as 2 pi, specify 2*np.pi (or leave as default).
        Can be a string or a dict containing key-value pairs of column name and period to be used for that column.

    Attributes
    ----------
    columns : str or list
        Columns to take the sine or cosine of.

    method : str or list
        The function to be calculated; either sin, cos or a list containing both.

    units : str or dict
        Which time unit the calculation is to be carried out on. Will take any of 'year', 'month',
        'day', 'hour', 'minute', 'second', 'microsecond'. Can be a string or a dict containing key-value pairs of column
        name and units to be used for that column.

    period : str, float or dict, default = 2*np.pi
        The period of the output in the units specified above. Can be a string or a dict containing key-value pairs of column
        name and units to be used for that column.
    """

    def __init__(
        self,
        columns: str | list[str],
        method: str | list[str],
        units: str | dict,
        period: int | float | dict = 2 * np.pi,
    ) -> None:
        super().__init__(columns, copy=True)

        if not isinstance(method, str) and not isinstance(method, list):
            msg = "{}: method must be a string or list but got {}".format(
                self.classname(),
                type(method),
            )
            raise TypeError(msg)

        if not isinstance(units, str) and not isinstance(units, dict):
            msg = "{}: units must be a string or dict but got {}".format(
                self.classname(),
                type(units),
            )
            raise TypeError(msg)

        if (
            (not isinstance(period, int))
            and (not isinstance(period, float))
            and (not isinstance(period, dict))
            or (isinstance(period, bool))
        ):
            msg = "{}: period must be an int, float or dict but got {}".format(
                self.classname(),
                type(period),
            )
            raise TypeError(msg)

        if isinstance(units, dict) and (
            not all(isinstance(item, str) for item in list(units.keys()))
            or not all(isinstance(item, str) for item in list(units.values()))
        ):
            msg = "{}: units dictionary key value pair must be strings but got keys: {} and values: {}".format(
                self.classname(),
                {type(k) for k in units},
                {type(v) for v in units.values()},
            )
            raise TypeError(msg)

        if isinstance(period, dict) and (
            not all(isinstance(item, str) for item in list(period.keys()))
            or (
                not all(isinstance(item, int) for item in list(period.values()))
                and not all(isinstance(item, float) for item in list(period.values()))
            )
            or any(isinstance(item, bool) for item in list(period.values()))
        ):
            msg = "{}: period dictionary key value pair must be str:int or str:float but got keys: {} and values: {}".format(
                self.classname(),
                {type(k) for k in period},
                {type(v) for v in period.values()},
            )
            raise TypeError(msg)

        valid_method_list = ["sin", "cos"]

        method_list = [method] if isinstance(method, str) else method

        for method in method_list:
            if method not in valid_method_list:
                msg = '{}: Invalid method {} supplied, should be "sin", "cos" or a list containing both'.format(
                    self.classname(),
                    method,
                )
                raise ValueError(msg)

        valid_unit_list = [
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "microsecond",
        ]

        if isinstance(units, dict):
            if not set(units.values()).issubset(valid_unit_list):
                msg = "{}: units dictionary values must be one of 'year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond' but got {}".format(
                    self.classname(),
                    set(units.values()),
                )
                raise ValueError(msg)
        elif units not in valid_unit_list:
            msg = "{}: Invalid units {} supplied, should be in {}".format(
                self.classname(),
                units,
                valid_unit_list,
            )
            raise ValueError(msg)

        self.method = method_list
        self.units = units
        self.period = period

        if isinstance(units, dict) and sorted(units.keys()) != sorted(self.columns):
            msg = "{}: unit dictionary keys must be the same as columns but got {}".format(
                self.classname(),
                set(units.keys()),
            )
            raise ValueError(msg)

        if isinstance(period, dict) and sorted(period.keys()) != sorted(self.columns):
            msg = "{}: period dictionary keys must be the same as columns but got {}".format(
                self.classname(),
                set(period.keys()),
            )
            raise ValueError(msg)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform - creates column containing sine or cosine of another datetime column.

        Which function is used is stored in the self.method attribute.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional columns added, these are named "<method>_<original_column>"
        """
        X = super().transform(X)

        temp = self.cast_columns_to_datetime(X)

        for column in self.columns:
            if not isinstance(self.units, dict):
                column_in_desired_unit = getattr(temp[column].dt, self.units)
                desired_units = self.units
            elif isinstance(self.units, dict):
                column_in_desired_unit = getattr(temp[column].dt, self.units[column])
                desired_units = self.units[column]
            if not isinstance(self.period, dict):
                desired_period = self.period
            elif isinstance(self.period, dict):
                desired_period = self.period[column]

            for method in self.method:
                new_column_name = f"{method}_{desired_period}_{desired_units}_{column}"

                X[new_column_name] = getattr(np, method)(
                    column_in_desired_unit * (2.0 * np.pi / desired_period),
                )

        return X



import re

import setuptools

with open("README.md") as fh:
    long_description = fh.read()

# get version from _version.py file, from below
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
VERSION_FILE = "tubular/_version.py"
VERSION_STR_RE = r"^__version__ = ['\"]([^'\"]*)['\"]"
with open(VERSION_FILE) as version_file:
    version_file_str = version_file.read()
    mo = re.search(VERSION_STR_RE, version_file_str, re.M)
    if mo:
        version = mo.group(1)
    else:
        msg = f"Unable to find version string in {VERSION_FILE}."
        raise RuntimeError(msg)


def list_reqs(fname: str = "requirements.txt") -> list:
    with open(fname) as fd:
        return fd.read().splitlines()


setuptools.setup(
    name="tubular",
    version=version,
    author="LV GI Data Science Team",
    author_email="#DataSciencePackages@lv.co.uk",
    description="Package to perform pre processing steps for machine learning models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=list_reqs(),
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
)



"""This module contains transformers that apply numeric functions."""

from __future__ import annotations

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import (
    MaxAbsScaler,
    MinMaxScaler,
    PolynomialFeatures,
    StandardScaler,
)

from tubular.base import BaseTransformer, DataFrameMethodTransformer


class LogTransformer(BaseTransformer):
    """Transformer to apply log transformation.

    Transformer has the option to add 1 to the columns to log and drop the
    original columns.

    Parameters
    ----------
    columns : None or str or list
        Columns to log transform.

    base : None or float/int
        Base for log transform. If None uses natural log.

    add_1 : bool
        Should a constant of 1 be added to the columns to be transformed prior to
        applying the log transform?

    drop : bool
        Should the original columns to be transformed be dropped after applying the
        log transform?

    suffix : str, default = '_log'
        The suffix to add onto the end of column names for new columns.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    add_1 : bool
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    drop : bool
        The name of the pandas.DataFrame method to call.

    suffix : str
        The suffix to add onto the end of column names for new columns.

    """

    def __init__(
        self,
        columns: str | list[str] | None,
        base: float | int | None = None,
        add_1: bool = False,
        drop: bool = True,
        suffix: str = "log",
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if base is not None:
            if not isinstance(base, (int, float)):
                msg = f"{self.classname()}: base should be numeric or None"
                raise ValueError(msg)
            if not base > 0:
                msg = f"{self.classname()}: base should be strictly positive"
                raise ValueError(msg)

        self.base = base
        self.add_1 = add_1
        self.drop = drop
        self.suffix = suffix

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Applies the log transform to the specified columns.

        If the drop attribute is True then the original columns are dropped. If
        the add_1 attribute is True then the original columns + 1 are logged.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe to be transformed.

        Returns
        -------
        X : pd.DataFrame
            The dataframe with the specified columns logged, optionally dropping the original
            columns if self.drop is True.

        """
        X = super().transform(X)

        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype,
            axis=0,
        )

        if not numeric_column_types.all():
            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index,
            )

            msg = f"{self.classname()}: The following columns are not numeric in X; {non_numeric_columns}"
            raise TypeError(msg)

        new_column_names = [f"{column}_{self.suffix}" for column in self.columns]

        if self.add_1:
            if (X[self.columns] <= -1).sum().sum() > 0:
                msg = f"{self.classname()}: values less than or equal to 0 in columns (after adding 1), make greater than 0 before using transform"
                raise ValueError(msg)

            if self.base is None:
                X[new_column_names] = np.log1p(X[self.columns])

            else:
                X[new_column_names] = np.log1p(X[self.columns]) / np.log(self.base)

        else:
            if (X[self.columns] <= 0).sum().sum() > 0:
                msg = f"{self.classname()}: values less than or equal to 0 in columns, make greater than 0 before using transform"
                raise ValueError(msg)

            if self.base is None:
                X[new_column_names] = np.log(X[self.columns])

            else:
                X[new_column_names] = np.log(X[self.columns]) / np.log(self.base)

        if self.drop:
            X = X.drop(self.columns, axis=1)

        return X


class CutTransformer(BaseTransformer):
    """Class to bin a column into discrete intervals.

    Class simply uses the [pd.cut](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.cut.html)
    method on the specified column.

    Parameters
    ----------
    column : str
        Name of the column to discretise.

    new_column_name : str
        Name given to the new discrete column.

    cut_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.cut method when it is called in transform.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init().

    """

    def __init__(
        self,
        column: str,
        new_column_name: str,
        cut_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if type(column) is not str:
            msg = f"{self.classname()}: column arg (name of column) should be a single str giving the column to discretise"
            raise TypeError(msg)

        if type(new_column_name) is not str:
            msg = f"{self.classname()}: new_column_name must be a str"
            raise TypeError(msg)

        if cut_kwargs is None:
            cut_kwargs = {}
        else:
            if type(cut_kwargs) is not dict:
                msg = f"{self.classname()}: cut_kwargs should be a dict but got type {type(cut_kwargs)}"
                raise TypeError(msg)

        for i, k in enumerate(cut_kwargs.keys()):
            if type(k) is not str:
                msg = f"{self.classname()}: unexpected type ({type(k)}) for cut_kwargs key in position {i}, must be str"
                raise TypeError(msg)

        self.cut_kwargs = cut_kwargs
        self.new_column_name = new_column_name

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

        super().__init__(columns=[column], **kwargs)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Discretise specified column using pd.cut.

        Parameters
        ----------
        X : pd.DataFrame
            Data with column to transform.

        """
        X = super().transform(X)

        if not pd.api.types.is_numeric_dtype(X[self.columns[0]]):
            msg = f"{self.classname()}: {self.columns[0]} should be a numeric dtype but got {X[self.columns[0]].dtype}"
            raise TypeError(msg)

        X[self.new_column_name] = pd.cut(
            X[self.columns[0]].to_numpy(),
            **self.cut_kwargs,
        )

        return X


class TwoColumnOperatorTransformer(DataFrameMethodTransformer):
    """This transformer applies a pandas.DataFrame method to two columns (add, sub, mul, div, mod, pow).

    Transformer assigns the output of the method to a new column. The method will be applied
    in the form (column 1)operator(column 2), so order matters (if the method does not commute). It is possible to
    supply other key word arguments to the transform method, which will be passed to the pandas.DataFrame method being called.

    Parameters
    ----------
    pd_method_name : str
        The name of the pandas.DataFrame method to be called.

    column1_name : str
        The name of the 1st column in the operation.

    column2_name : str
        The name of the 2nd column in the operation.

    new_column_name : str
        The name of the new column that the output is assigned to.

    pd_method_kwargs : dict, default =  {'axis':0}
        Dictionary of method kwargs to be passed to pandas.DataFrame method. Must contain an entry for axis, set to either 1 or 0.

    **kwargs :
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    pd_method_name : str
        The name of the pandas.DataFrame method to be called.

    columns : list
        list containing two string items: [column1_name, column2_name] The first will be operated upon by the
        chosen pandas method using the second.

    column2_name : str
        The name of the 2nd column in the operation.

    new_column_name : str
        The name of the new column that the output is assigned to.

    pd_method_kwargs : dict
        Dictionary of method kwargs to be passed to pandas.DataFrame method.

    """

    def __init__(
        self,
        pd_method_name: str,
        columns: list[str],
        new_column_name: str,
        pd_method_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        """Performs input checks not done in either DataFrameMethodTransformer.__init__ or BaseTransformer.__init__."""
        if pd_method_kwargs is None:
            pd_method_kwargs = {"axis": 0}
        else:
            if "axis" not in pd_method_kwargs.keys():
                msg = f'{self.classname()}: pd_method_kwargs must contain an entry "axis" set to 0 or 1'
                raise ValueError(msg)
            if pd_method_kwargs["axis"] not in [0, 1]:
                msg = f"{self.classname()}: pd_method_kwargs 'axis' must be 0 or 1"
                raise ValueError(msg)

        if type(columns) is not list and len(columns) != 2:
            msg = f"{self.classname()}: columns must be a list containing two column names but got {columns}"
            raise ValueError(msg)

        self.column1_name = columns[0]
        self.column2_name = columns[1]

        # call DataFrameMethodTransformer.__init__
        # This class will inherit all the below attributes from DataFrameMethodTransformer
        super().__init__(
            new_column_name=new_column_name,
            pd_method_name=pd_method_name,
            columns=columns,
            pd_method_kwargs=pd_method_kwargs,
            **kwargs,
        )

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform input data by applying the chosen method to the two specified columns.

        Args:
        ----
            X (pd.DataFrame): Data to transform.

        Returns:
        -------
            pd.DataFrame: Input X with an additional column.
        """
        # call BaseTransformer.transform
        X = super(DataFrameMethodTransformer, self).transform(X)

        is_numeric = X[self.columns].apply(pd.api.types.is_numeric_dtype, axis=0)

        if not is_numeric.all():
            msg = f"{self.classname()}: input columns in X must contain only numeric values"
            raise TypeError(msg)

        X[self.new_column_name] = getattr(X[[self.column1_name]], self.pd_method_name)(
            X[self.column2_name],
            **self.pd_method_kwargs,
        )

        return X


class ScalingTransformer(BaseTransformer):
    """Transformer to perform scaling of numeric columns.

    Transformer can apply min max scaling, max absolute scaling or standardisation (subtract mean and divide by std).
    The transformer uses the appropriate sklearn.preprocessing scaler.

    Parameters
    ----------
    columns : str, list or None
        Name of the columns to apply scaling to.

    scaler_type : str
        Type of scaler to use, must be one of 'min_max', 'max_abs' or 'standard'. The corresponding
        sklearn.preprocessing scaler used in each case is MinMaxScaler, MaxAbsScaler or StandardScaler.

    scaler_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the scaler object when it is initialised.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init().

    """

    def __init__(
        self,
        columns: str | list[str] | None,
        scaler_type: str,
        scaler_kwargs: dict[str, object] | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        if scaler_kwargs is None:
            scaler_kwargs = {}
        else:
            if type(scaler_kwargs) is not dict:
                msg = f"{self.classname()}: scaler_kwargs should be a dict but got type {type(scaler_kwargs)}"
                raise TypeError(msg)

        for i, k in enumerate(scaler_kwargs.keys()):
            if type(k) is not str:
                msg = f"{self.classname()}: unexpected type ({type(k)}) for scaler_kwargs key in position {i}, must be str"
                raise TypeError(msg)

        allowed_scaler_values = ["min_max", "max_abs", "standard"]

        if scaler_type not in allowed_scaler_values:
            msg = f"{self.classname()}: scaler_type should be one of; {allowed_scaler_values}"
            raise ValueError(msg)

        if scaler_type == "min_max":
            self.scaler = MinMaxScaler(**scaler_kwargs)

        elif scaler_type == "max_abs":
            self.scaler = MaxAbsScaler(**scaler_kwargs)

        elif scaler_type == "standard":
            self.scaler = StandardScaler(**scaler_kwargs)

        # This attribute is not for use in any method
        # Here only as a fix to allow string representation of transformer.
        self.scaler_kwargs = scaler_kwargs
        self.scaler_type = scaler_type

        super().__init__(columns=columns, **kwargs)

    def check_numeric_columns(self, X: pd.DataFrame) -> pd.DataFrame:
        """Method to check all columns (specicifed in self.columns) in X are all numeric.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing columns to check.

        """
        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype,
            axis=0,
        )

        if not numeric_column_types.all():
            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index,
            )

            msg = f"{self.classname()}: The following columns are not numeric in X; {non_numeric_columns}"
            raise TypeError(msg)

        return X

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Fit scaler to input data.

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe with columns to learn scaling values from.

        y : None
            Required for pipeline.

        """
        super().fit(X, y)

        X = self.check_numeric_columns(X)

        self.scaler.fit(X[self.columns])

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform input data X with fitted scaler.

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing columns to be scaled.

        Returns
        -------
        X : pd.DataFrame
            Input X with columns scaled.

        """
        X = super().transform(X)

        X = self.check_numeric_columns(X)

        X[self.columns] = self.scaler.transform(X[self.columns])

        return X


class InteractionTransformer(BaseTransformer):
    """Transformer that generates interaction features.
    Transformer generates a new column  for all combinations from the selected columns up to the maximum degree
    provided. (For sklearn version higher than 1.0.0>, only interaction of a degree higher or equal to the minimum
    degree would be computed).
    Each interaction column consists of the product of the specific combination of columns.
    Ex: with 3 columns provided ["a","b","c"], if max degree is 3, the total possible combinations are :
    - of degree 1 : ["a","b","c"]
    - of degree 2 : ["a b","b c","a c"]
    - of degree 3 : ["a b c"].

    Parameters
    ----------
        columns : None or list or str
            Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
            in columns is saved in the columns attribute on the object. Note this has no default value so
            the user has to specify the columns when initialising the transformer. This is avoid likely
            when the user forget to set columns, in this case all columns would be picked up when super
            transform runs.
        min_degree : int
            minimum degree of interaction features to be considered. For example if min_degree=3, only interaction
            columns from at least 3 columns would be generated. NB- only applies if sklearn version is 1.0.0>=
        max_degree : int
            maximum degree of interaction features to be considered. For example if max_degree=3, only interaction
            columns from up to 3 columns would be generated.


    Attributes
    ----------
        min_degree : int
            minimum degree of interaction features to be considered
        max_degree : int
            maximum degree of interaction features to be considered
        nb_features_to_interact : int
            number of selected columns from which interactions should be computed. (=len(columns))
        nb_combinations : int
            number of new interaction features
        interaction_colname : list
            names of each new interaction feature. The name of an interaction feature is the combinations of previous
            column names joined with a whitespace. Interaction feature of ["col1","col2","col3] would be "col1 col2 col3".
        nb_feature_out : int
            number of total columns of transformed dataset, including new interaction features

    """

    def __init__(
        self,
        columns: str | list[str] | None,
        min_degree: int = 2,
        max_degree: int = 2,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if len(columns) < 2:
            msg = f"{self.classname()}: number of columns must be equal or greater than 2, got {str(len(columns))} column."
            raise ValueError(msg)

        if type(min_degree) is int:
            if min_degree < 2:
                msg = f"{self.classname()}: min_degree must be equal or greater than 2, got {str(min_degree)}"
                raise ValueError(msg)
            self.min_degree = min_degree
        else:
            msg = f"{self.classname()}: unexpected type ({type(min_degree)}) for min_degree, must be int"
            raise TypeError(msg)
        if type(max_degree) is int:
            if min_degree > max_degree:
                msg = f"{self.classname()}: max_degree must be equal or greater than min_degree"
                raise ValueError(msg)
            self.max_degree = max_degree
            if max_degree > len(columns):
                msg = f"{self.classname()}: max_degree must be equal or lower than number of columns"
                raise ValueError(msg)
            self.max_degree = max_degree
            if max_degree > len(columns):
                msg = f"{self.classname()}: max_degree must be equal or lower than number of columns"
                raise ValueError(msg)
            self.max_degree = max_degree
        else:
            msg = f"{self.classname()}: unexpected type ({type(max_degree)}) for max_degree, must be int"
            raise TypeError(msg)

        self.nb_features_to_interact = len(self.columns)
        self.nb_combinations = -1
        self.interaction_colname = []
        self.nb_feature_out = -1

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Generate from input pandas DataFrame (X) new interaction features using the "product" pandas.DataFrame method
         and add this column or columns in X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column or columns (self.interaction_colname) added. These contain the output of
            running the  product pandas DataFrame method on identified combinations.

        """
        X = super().transform(X)

        try:
            interaction_combination_index = PolynomialFeatures._combinations(
                n_features=self.nb_features_to_interact,
                min_degree=self.min_degree,
                max_degree=self.max_degree,
                interaction_only=True,
                include_bias=False,
            )
        except TypeError as err:
            if (
                str(err)
                == "_combinations() got an unexpected keyword argument 'min_degree'"
            ):
                interaction_combination_index = PolynomialFeatures._combinations(
                    n_features=self.nb_features_to_interact,
                    degree=self.max_degree,
                    interaction_only=True,
                    include_bias=False,
                )
            else:
                raise err

        interaction_combination_colname = [
            [self.columns[col_idx] for col_idx in interaction_combination]
            for interaction_combination in interaction_combination_index
        ]
        self.nb_combinations = len(interaction_combination_colname)
        self.nb_feature_out = self.nb_combinations + len(X)

        self.interaction_colname = [
            " ".join(interaction_combination)
            for interaction_combination in interaction_combination_colname
        ]

        for inter_idx in range(len(interaction_combination_colname)):
            X[self.interaction_colname[inter_idx]] = X[
                interaction_combination_colname[inter_idx]
            ].product(axis=1, skipna=False)

        return X


class PCATransformer(BaseTransformer):
    """Transformer that generates variables using Principal component analysis (PCA).
    Linear dimensionality reduction using Singular Value Decomposition of the
    data to project it to a lower dimensional space.

    It is based on sklearn class sklearn.decomposition.PCA

    Parameters
    ----------
        columns : None or list or str
            Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
            in columns is saved in the columns attribute on the object. Note this has no default value so
            the user has to specify the columns when initialising the transformer. When the user forget to set columns,
            all columns would be picked up when super transform runs.
        n_components : int, float or 'mle', default=None
            Number of components to keep.
            if n_components is not set all components are kept::
                n_components == min(n_samples, n_features)
            If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
            MLE is used to guess the dimension. Use of ``n_components == 'mle'``
            will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.
            If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
            number of components such that the amount of variance that needs to be
            explained is greater than the percentage specified by n_components.
            If ``svd_solver == 'arpack'``, the number of components must be
             strictly less than the minimum of n_features and n_samples.
            Hence, the None case results in::
                n_components == min(n_samples, n_features) - 1   svd_solver='auto', tol=0.0,  n_oversamples=10, random_state=None
        svd_solver : {'auto', 'full', 'arpack', 'randomized'}, default='auto'
            If auto :
                The solver is selected by a default policy based on `X.shape` and
                `n_components`: if the input data is larger than 500x500 and the
                number of components to extract is lower than 80% of the smallest
                dimension of the data, then the more efficient 'randomized'
                method is enabled. Otherwise the exact full SVD is computed and
                optionally truncated afterwards.
            If full :
                run exact full SVD calling the standard LAPACK solver via
                `scipy.linalg.svd` and select the components by postprocessing
            If arpack :
                run SVD truncated to n_components calling ARPACK solver via
                `scipy.sparse.linalg.svds`. It requires strictly
                0 < n_components < min(X.shape)
            If randomized :
                run randomized SVD by the method of Halko et al.
            .. sklearn versionadded:: 0.18.0

        random_state : int, RandomState instance or None, default=None
            Used when the 'arpack' or 'randomized' solvers are used. Pass an int
            for reproducible results across multiple function calls.
            .. sklearn versionadded:: 0.18.0
        pca_column_prefix : str, prefix added to each the n components features generated. Default is "pca_"
            example: if n_components = 3, new columns would be 'pca_0','pca_1','pca_2'.

    Attributes
    ----------
    pca : PCA class from sklearn.decomposition
    n_components_ : int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features and n_samples
        if n_components is None.
    feature_names_out: list or None
        list of feature name representing the new dimensions.


    """

    def __init__(
        self,
        columns: str | list[str] | None,
        n_components: int = 2,
        svd_solver: str = "auto",
        random_state: int | np.random.RandomState = None,
        pca_column_prefix: str = "pca_",
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if type(n_components) is int:
            if n_components < 1:
                msg = f"{self.classname()}:n_components must be strictly positive got {str(n_components)}"
                raise ValueError(msg)
            self.n_components = n_components
        elif type(n_components) is float:
            if 0 < n_components < 1:
                self.n_components = n_components
            else:
                msg = f"{self.classname()}:n_components must be strictly positive and must be of type int when greater than or equal to 1. Got {str(n_components)}"
                raise ValueError(msg)

        else:
            if n_components == "mle":
                self.n_components = n_components
            else:
                msg = f"{self.classname()}:unexpected type {type(n_components)} for n_components, must be int, float (0-1) or equal to 'mle'."
                raise TypeError(msg)

        if type(svd_solver) is str:
            if svd_solver not in ["auto", "full", "arpack", "randomized"]:
                msg = f"{self.classname()}:svd_solver {svd_solver} is unknown. Please select among 'auto', 'full', 'arpack', 'randomized'."
                raise ValueError(msg)
            self.svd_solver = svd_solver
        else:
            msg = f"{self.classname()}:unexpected type {type(svd_solver)} for svd_solver, must be str"
            raise TypeError(msg)

        if type(random_state) is int or random_state is None:
            self.random_state = random_state
        else:
            msg = f"{self.classname()}:unexpected type {type(random_state)} for random_state, must be int or None."
            raise TypeError(msg)

        if (svd_solver == "arpack") and (n_components == "mle"):
            msg = f"{self.classname()}: n_components='mle' cannot be a string with svd_solver='arpack'"
            raise ValueError(msg)
        if (svd_solver in ["randomized", "arpack"]) and (type(n_components) is float):
            msg = f"{self.classname()}: n_components {n_components} cannot be a float with svd_solver='{svd_solver}'"
            raise TypeError(msg)

        if type(pca_column_prefix) is str:
            self.pca_column_prefix = pca_column_prefix
        else:
            msg = f"{self.classname()}:unexpected type {type(pca_column_prefix)} for pca_column_prefix, must be str"
            raise TypeError(msg)

        self.pca = PCA(
            n_components=self.n_components,
            svd_solver=self.svd_solver,
            random_state=self.random_state,
        )

        self.pca_column_prefix = pca_column_prefix
        self.feature_names_out = None
        self.n_components_ = None

    def check_numeric_columns(self, X: pd.DataFrame) -> pd.DataFrame:
        """Method to check all columns (specicifed in self.columns) in X are all numeric.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing columns to check.

        """
        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype,
            axis=0,
        )

        if not numeric_column_types.all():
            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index,
            )

            msg = f"{self.classname()}: The following columns are not numeric in X; {non_numeric_columns}"
            raise TypeError(msg)

        return X

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Fit PCA to input data.

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe with columns to learn scaling values from.

        y : None
            Required for pipeline.

        """
        super().fit(X, y)

        X = self.check_numeric_columns(X)

        if self.n_components != "mle":
            if 0 < self.n_components <= min(X[self.columns].shape):
                pass
            else:
                msg = f"{self.classname()}: n_components {self.n_components} must be between 1 and min(n_samples {X[self.columns].shape[0]}, n_features {X[self.columns].shape[1]}) is {min(X[self.columns].shape)} with svd_solver '{self.svd_solver}'"
                raise ValueError(msg)

        self.pca.fit(X[self.columns])
        self.n_components_ = self.pca.n_components_
        self.feature_names_out = [
            self.pca_column_prefix + str(i) for i in range(self.n_components_)
        ]

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Generate from input pandas DataFrame (X) PCA features and add this column or columns in X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column or columns (self.interaction_colname) added. These contain the output of
            running the  product pandas DataFrame method on identified combinations.
        """
        X = super().transform(X)
        X = self.check_numeric_columns(X)
        X[self.feature_names_out] = self.pca.transform(X[self.columns])

        return X



"""This module contains transformers that other transformers in the package inherit
from. These transformers contain key checks to be applied in all cases.
"""

from __future__ import annotations

import warnings

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from tubular._version import __version__


class BaseTransformer(TransformerMixin, BaseEstimator):
    """Base tranformer class which all other transformers in the package inherit from.

    Provides fit and transform methods (required by sklearn transformers), simple input checking
    and functionality to copy X prior to transform.

    Parameters
    ----------
    columns : None or list or str, default = None
        Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object.

    copy : bool, default = True
        Should X be copied before tansforms are applied?

    verbose : bool, default = False
        Should statements be printed when methods are run?

    Attributes
    ----------
    columns : list or None
        Either a list of str values giving which columns in a input pandas.DataFrame the transformer
        will be applied to - or None.

    copy : bool
        Should X be copied before tansforms are applied?

    verbose : bool
        Print statements to show which methods are being run or not.

    version_ : str
        Version number (__version__ attribute from _version.py).

    """

    def classname(self) -> str:
        """Method that returns the name of the current class when called."""
        return type(self).__name__

    def __init__(
        self,
        columns: list[str] | str = None,
        copy: bool = True,
        verbose: bool = False,
    ) -> None:
        self.version_ = __version__

        if not isinstance(verbose, bool):
            msg = f"{self.classname()}: verbose must be a bool"
            raise TypeError(msg)

        self.verbose = verbose

        if self.verbose:
            print("BaseTransformer.__init__() called")

        if columns is None:
            self.columns = None

        else:
            # make sure columns is a single str or list of strs
            if isinstance(columns, str):
                self.columns = [columns]

            elif isinstance(columns, list):
                if not len(columns) > 0:
                    msg = f"{self.classname()}: columns has no values"
                    raise ValueError(msg)

                for c in columns:
                    if not isinstance(c, str):
                        msg = f"{self.classname()}: each element of columns should be a single (string) column name"
                        raise TypeError(msg)

                self.columns = columns

            else:
                msg = f"{self.classname()}: columns must be a string or list with the columns to be pre-processed (if specified)"
                raise TypeError(msg)

        if not isinstance(copy, bool):
            msg = f"{self.classname()}: copy must be a bool"
            raise TypeError(msg)

        self.copy = copy

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> BaseTransformer:
        """Base transformer fit method, checks X and y types. Currently only pandas DataFrames are allowed for X
        and DataFrames or Series for y.

        Fit calls the columns_set_or_check method which will set the columns attribute to all columns in X, if it
        is None.

        Parameters
        ----------
        X : pd.DataFrame
            Data to fit the transformer on.

        y : None or pd.DataFrame or pd.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        if self.verbose:
            print("BaseTransformer.fit() called")

        self.columns_set_or_check(X)

        if not X.shape[0] > 0:
            msg = f"{self.classname()}: X has no rows; {X.shape}"
            raise ValueError(msg)

        if y is not None:
            if not isinstance(y, pd.Series):
                msg = (
                    f"{self.classname()}: unexpected type for y, should be a pd.Series"
                )
                raise TypeError(msg)

            if not y.shape[0] > 0:
                msg = f"{self.classname()}: y is empty; {y.shape}"
                raise ValueError(msg)

        return self

    def _combine_X_y(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """Combine X and y by adding a new column with the values of y to a copy of X.

        The new column response column will be called `_temporary_response`.

        This method can be used by transformers that need to use the response, y, together
        with the explanatory variables, X, in their `fit` methods.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing explanatory variables.

        y : pd.Series
            Response variable.

        """
        if not isinstance(X, pd.DataFrame):
            msg = f"{self.classname()}: X should be a pd.DataFrame"
            raise TypeError(msg)

        if not isinstance(y, pd.Series):
            msg = f"{self.classname()}: y should be a pd.Series"
            raise TypeError(msg)

        if X.shape[0] != y.shape[0]:
            msg = f"{self.classname()}: X and y have different numbers of rows ({X.shape[0]} vs {y.shape[0]})"
            raise ValueError(msg)

        if not (X.index == y.index).all():
            warnings.warn(
                f"{self.classname()}: X and y do not have equal indexes",
                stacklevel=2,
            )

        X_y = X.copy()

        X_y["_temporary_response"] = y.to_numpy()

        return X_y

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Base transformer transform method; checks X type (pandas DataFrame only) and copies data if requested.

        Transform calls the columns_check method which will check columns in columns attribute are in X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform with the transformer.

        Returns
        -------
        X : pd.DataFrame
            Input X, copied if specified by user.

        """
        self.columns_check(X)

        if self.verbose:
            print("BaseTransformer.transform() called")

        if self.copy:
            X = X.copy()

        if not X.shape[0] > 0:
            msg = f"{self.classname()}: X has no rows; {X.shape}"
            raise ValueError(msg)

        return X

    def check_is_fitted(self, attribute: str) -> None:
        """Check if particular attributes are on the object. This is useful to do before running transform to avoid
        trying to transform data without first running the fit method.

        Wrapper for utils.validation.check_is_fitted function.

        Parameters
        ----------
        attributes : List
            List of str values giving names of attribute to check exist on self.

        """
        check_is_fitted(self, attribute)

    def columns_check(self, X: pd.DataFrame) -> None:
        """Method to check that the columns attribute is set and all values are present in X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to check columns are in.

        """
        if not isinstance(X, pd.DataFrame):
            msg = f"{self.classname()}: X should be a pd.DataFrame"
            raise TypeError(msg)

        if self.columns is None:
            msg = f"{self.classname()}: columns not set"
            raise ValueError(msg)

        if not isinstance(self.columns, list):
            msg = f"{self.classname()}: self.columns should be a list"
            raise TypeError(msg)

        for c in self.columns:
            if c not in X.columns.to_numpy():
                raise ValueError(f"{self.classname()}: variable " + c + " is not in X")

    def columns_set_or_check(self, X: pd.DataFrame) -> None:
        """Function to check or set columns attribute.

        If the columns attribute is None then set it to all columns in X. Otherwise run the columns_check method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to check columns are in.

        """
        if not isinstance(X, pd.DataFrame):
            msg = f"{self.classname()}: X should be a pd.DataFrame"
            raise TypeError(msg)

        if self.columns is None:
            self.columns = list(X.columns.values)

        else:
            self.columns_check(X)

    @staticmethod
    def check_weights_column(X: pd.DataFrame, weights_column: str) -> None:
        """Helper method for validating weights column in dataframe.

        Args:
        ----
            X (pd.DataFrame): df containing weight column
            weights_column (str): name of weight column

        """
        if weights_column is not None:
            # check if given weight is in columns
            if weights_column not in X.columns:
                msg = f"weight col ({weights_column}) is not present in columns of data"
                raise ValueError(msg)

            # check weight is numeric

            if not pd.api.types.is_numeric_dtype(X[weights_column]):
                msg = "weight column must be numeric."
                raise ValueError(msg)

            # check weight is positive

            if (X[weights_column] < 0).sum() != 0:
                msg = "weight column must be positive"
                raise ValueError(msg)

            # check weight non-null
            if X[weights_column].isna().sum() != 0:
                msg = "weight column must be non-null"
                raise ValueError(msg)


class DataFrameMethodTransformer(BaseTransformer):
    """Tranformer that applies a pandas.DataFrame method.

    Transformer assigns the output of the method to a new column or columns. It is possible to
    supply other key word arguments to the transform method, which will be passed to the
    pandas.DataFrame method being called.

    Be aware it is possible to supply incompatible arguments to init that will only be
    identified when transform is run. This is because there are many combinations of method, input
    and output sizes. Additionally some methods may only work as expected when called in
    transform with specific key word arguments.

    Parameters
    ----------
    new_column_name : str or list of str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    columns : None or list or str
        Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object. Note this has no default value so
        the user has to specify the columns when initialising the transformer. This is avoid likely
        when the user forget to set columns, in this case all columns would be picked up when super
        transform runs.

    pd_method_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.DataFrame method when it is called.

    drop_original : bool, default = False
        Should original columns be dropped?

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    new_column_name : str or list of str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    """

    def __init__(
        self,
        new_column_name: list[str] | str,
        pd_method_name: str,
        columns: list[str] | str | None,
        pd_method_kwargs: dict[str, object] | None = None,
        drop_original: bool = False,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if type(new_column_name) is list:
            for i, item in enumerate(new_column_name):
                if type(item) is not str:
                    msg = f"{self.classname()}: if new_column_name is a list, all elements must be strings but got {type(item)} in position {i}"
                    raise TypeError(msg)

        elif type(new_column_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(new_column_name)}) for new_column_name, must be str or list of strings"
            raise TypeError(msg)

        if type(pd_method_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(pd_method_name)}) for pd_method_name, expecting str"
            raise TypeError(msg)

        if pd_method_kwargs is None:
            pd_method_kwargs = {}
        else:
            if type(pd_method_kwargs) is not dict:
                msg = f"{self.classname()}: pd_method_kwargs should be a dict but got type {type(pd_method_kwargs)}"
                raise TypeError(msg)

            for i, k in enumerate(pd_method_kwargs.keys()):
                if type(k) is not str:
                    msg = f"{self.classname()}: unexpected type ({type(k)}) for pd_method_kwargs key in position {i}, must be str"
                    raise TypeError(msg)

        if type(drop_original) is not bool:
            msg = f"{self.classname()}: unexpected type ({type(drop_original)}) for drop_original, expecting bool"
            raise TypeError(msg)

        self.new_column_name = new_column_name
        self.pd_method_name = pd_method_name
        self.pd_method_kwargs = pd_method_kwargs
        self.drop_original = drop_original

        try:
            df = pd.DataFrame()
            getattr(df, pd_method_name)

        except Exception as err:
            msg = f'{self.classname()}: error accessing "{pd_method_name}" method on pd.DataFrame object - pd_method_name should be a pd.DataFrame method'
            raise AttributeError(msg) from err

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform input pandas DataFrame (X) using the given pandas.DataFrame method and assign the output
        back to column or columns in X.

        Any keyword arguments set in the pd_method_kwargs attribute are passed onto the pandas DataFrame method when calling it.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column or columns (self.new_column_name) added. These contain the output of
            running the pandas DataFrame method.

        """
        X = super().transform(X)

        X[self.new_column_name] = getattr(X[self.columns], self.pd_method_name)(
            **self.pd_method_kwargs,
        )

        if self.drop_original:
            X = X.drop(self.columns, axis=1)

        return X
